# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Toolbar Container.
"""

# Standard library imports
from collections import OrderedDict
import logging
from typing import Dict, List, Optional, Tuple, Union

# Third party imports
from qtpy.QtCore import QSize, Slot
from qtpy.QtWidgets import QAction, QWidget
from qtpy import PYSIDE2, PYSIDE6

# Local imports
from spyder.api.exceptions import SpyderAPIError
from spyder.api.translations import _
from spyder.api.utils import get_class_values
from spyder.api.widgets.main_container import PluginMainContainer
from spyder.api.widgets.toolbars import ApplicationToolbar
from spyder.config.base import DEV
from spyder.plugins.toolbar.api import ApplicationToolbars
from spyder.utils.qthelpers import SpyderAction
from spyder.utils.registries import ACTION_REGISTRY, TOOLBAR_REGISTRY


# Logging
logger = logging.getLogger(__name__)

# Type annotations
ToolbarItem = Union[SpyderAction, QWidget]
ItemInfo = Tuple[ToolbarItem, Optional[str], Optional[str], Optional[str]]


class ToolbarMenus:
    ToolbarsMenu = "toolbars_menu"


class ToolbarsMenuSections:
    Main = "main_section"
    Secondary = "secondary_section"


class ToolbarActions:
    ShowToolbars = "show toolbars"


class QActionID(QAction):
    """Wrapper class around QAction that allows to set/get an identifier."""
    @property
    def action_id(self):
        return self._action_id

    @action_id.setter
    def action_id(self, act):
        self._action_id = act


class ToolbarContainer(PluginMainContainer):

    def __init__(self, name, plugin, parent=None):
        super().__init__(name, plugin, parent=parent)

        self._APPLICATION_TOOLBARS = OrderedDict()
        self._ADDED_TOOLBARS = OrderedDict()
        self._toolbarslist: list[ApplicationToolbar] = []
        self._visible_toolbars: list[ApplicationToolbar] = []
        self._ITEMS_QUEUE: Dict[str, List[ItemInfo]] = {}

    # ---- Private Methods
    # ------------------------------------------------------------------------
    def _save_visible_toolbars(self):
        """Save the name of the visible toolbars in the options."""
        toolbars = []
        for toolbar in self._visible_toolbars:
            toolbars.append(toolbar.objectName())

        self.set_conf('last_visible_toolbars', toolbars)

    def _set_visible_toolbars(self):
        """Set the current visible toolbars in an attribute."""
        toolbars = []
        for toolbar in self._toolbarslist:
            if (
                toolbar.toggleViewAction().isChecked()
                and toolbar not in toolbars
            ):
                toolbars.append(toolbar)

        self._visible_toolbars = toolbars

    @Slot()
    def _show_toolbars(self):
        """Show/Hide toolbars."""
        value = not self.get_conf("toolbars_visible")
        self.set_conf("toolbars_visible", value)
        if value:
            self._save_visible_toolbars()
        else:
            self._set_visible_toolbars()

        for toolbar in self._visible_toolbars:
            toolbar.setVisible(value)

        self.update_actions()

    def _add_missing_toolbar_elements(self, toolbar, toolbar_id):
        if toolbar_id in self._ITEMS_QUEUE:
            pending_items = self._ITEMS_QUEUE.pop(toolbar_id)
            for item, section, before, before_section in pending_items:
                toolbar.add_item(item, section=section, before=before,
                                 before_section=before_section)

    # ---- PluginMainContainer API
    # ------------------------------------------------------------------------
    def setup(self):
        self.show_toolbars_action = self.create_action(
            ToolbarActions.ShowToolbars,
            text=_("Show toolbars"),
            triggered=self._show_toolbars
        )

        self.toolbars_menu = self.create_menu(
            ToolbarMenus.ToolbarsMenu,
            _("Toolbars"),
        )

    def update_actions(self):
        visible_toolbars = self.get_conf("toolbars_visible")
        if visible_toolbars:
            text = _("Hide toolbars")
            tip = _("Hide toolbars")
        else:
            text = _("Show toolbars")
            tip = _("Show toolbars")

        self.show_toolbars_action.setText(text)
        self.show_toolbars_action.setToolTip(tip)
        self.toolbars_menu.setEnabled(visible_toolbars)

    # ---- Public API
    # ------------------------------------------------------------------------
    def create_application_toolbar(
        self,
        toolbar_id: str,
        title: str
    ) -> ApplicationToolbar:
        """
        Create an application toolbar and add it to the main window.

        Parameters
        ----------
        toolbar_id: str
            The toolbar unique identifier string.
        title: str
            The localized toolbar title to be displayed.

        Returns
        -------
        spyder.api.widgets.toolbar.ApplicationToolbar
            The created application toolbar.
        """
        if toolbar_id in self._APPLICATION_TOOLBARS:
            raise SpyderAPIError(
                'Toolbar with ID "{}" already added!'.format(toolbar_id)
            )

        toolbar = ApplicationToolbar(self, toolbar_id, title)
        self._add_missing_toolbar_elements(toolbar, toolbar_id)
        return toolbar

    def add_application_toolbar(self, toolbar, mainwindow=None):
        """
        Add toolbar to application toolbars.

        Parameters
        ----------
        toolbar: spyder.api.widgets.toolbars.ApplicationToolbar
            The application toolbar to add to the `mainwindow`.
        mainwindow: QMainWindow
            The main application window.
        """
        # Check toolbar class
        if not isinstance(toolbar, ApplicationToolbar):
            raise SpyderAPIError(
                'Any toolbar must subclass ApplicationToolbar!'
            )

        # Check ID
        toolbar_id = toolbar.ID
        if toolbar_id is None:
            raise SpyderAPIError(
                f"Toolbar `{repr(toolbar)}` doesn't have an identifier!"
            )

        if toolbar_id in self._ADDED_TOOLBARS:
            raise SpyderAPIError(
                'Toolbar with ID "{}" already added!'.format(toolbar_id))

        # Add toolbar to registry and add it to the app toolbars dict
        TOOLBAR_REGISTRY.register_reference(
            toolbar, toolbar_id, self.PLUGIN_NAME, self.CONTEXT_NAME
        )
        self._APPLICATION_TOOLBARS[toolbar_id] = toolbar

        # TODO: Make the icon size adjustable in Preferences later on.
        iconsize = 24
        toolbar.setIconSize(QSize(iconsize, iconsize))
        toolbar.setObjectName(toolbar_id)

        self._ADDED_TOOLBARS[toolbar_id] = toolbar
        self._toolbarslist.append(toolbar)

        if mainwindow:
            mainwindow.addToolBar(toolbar)

        self._add_missing_toolbar_elements(toolbar, toolbar_id)

    def remove_application_toolbar(self, toolbar_id: str, mainwindow=None):
        """
        Remove toolbar from application toolbars.

        Parameters
        ----------
        toolbar: str
            The application toolbar to remove from the `mainwindow`.
        mainwindow: QMainWindow
            The main application window.
        """

        if toolbar_id not in self._ADDED_TOOLBARS:
            raise SpyderAPIError(
                'Toolbar with ID "{}" is not in the main window'.format(
                    toolbar_id))

        toolbar = self._ADDED_TOOLBARS.pop(toolbar_id)
        self._toolbarslist.remove(toolbar)

        if mainwindow:
            mainwindow.removeToolBar(toolbar)

    def add_item_to_application_toolbar(
        self,
        item: ToolbarItem,
        toolbar_id: Optional[str] = None,
        section: Optional[str] = None,
        before: Optional[str] = None,
        before_section: Optional[str] = None,
        omit_id: bool = False
    ):
        """
        Add action or widget `item` to given application toolbar `section`.

        Parameters
        ----------
        item: SpyderAction or QWidget
            The item to add to the `toolbar`.
        toolbar_id: str or None
            The application toolbar unique string identifier.
        section: str or None
            The section id in which to insert the `item` on the `toolbar`.
        before: str or None
            Make the item appear before another given item.
        before_section: str or None
            Make the item defined section appear before another given section
            (the section must be already defined).
        omit_id: bool
            If True, then the toolbar will check if the item to add declares an
            id, False otherwise. This flag exists only for items added on
            Spyder 4 plugins. Default: False
        """
        if toolbar_id not in self._APPLICATION_TOOLBARS:
            pending_items = self._ITEMS_QUEUE.get(toolbar_id, [])
            pending_items.append((item, section, before, before_section))
            self._ITEMS_QUEUE[toolbar_id] = pending_items
        else:
            toolbar = self.get_application_toolbar(toolbar_id)
            toolbar.add_item(item, section=section, before=before,
                             before_section=before_section, omit_id=omit_id)

    def remove_item_from_application_toolbar(
        self,
        item_id: str,
        toolbar_id: Optional[str] = None
    ):
        """
        Remove action or widget from given application toolbar by id.

        Parameters
        ----------
        item: str
            The item to remove from the `toolbar`.
        toolbar_id: str or None
            The application toolbar unique string identifier.
        """
        if toolbar_id not in self._APPLICATION_TOOLBARS:
            raise SpyderAPIError(
                '{} is not a valid toolbar_id'.format(toolbar_id))

        toolbar = self.get_application_toolbar(toolbar_id)
        toolbar.remove_item(item_id)

    def get_application_toolbar(self, toolbar_id: str) -> ApplicationToolbar:
        """
        Return an application toolbar by toolbar_id.

        Parameters
        ----------
        toolbar_id: str
            The toolbar unique string identifier.

        Returns
        -------
        spyder.api.widgets.toolbars.ApplicationToolbar
            The application toolbar.
        """
        if toolbar_id not in self._APPLICATION_TOOLBARS:
            raise SpyderAPIError(
                'Application toolbar "{0}" not found! '
                'Available toolbars are: {1}'.format(
                    toolbar_id,
                    list(self._APPLICATION_TOOLBARS.keys())
                )
            )

        return self._APPLICATION_TOOLBARS[toolbar_id]

    def get_application_toolbars(self) -> Dict[str, ApplicationToolbar]:
        """
        Return all created application toolbars.

        Returns
        -------
        dict
            The dict of all the added application toolbars.
        """
        return self._APPLICATION_TOOLBARS

    def load_application_toolbars(self):
        """Load application toolbars at startup."""
        app_toolbars = self.get_application_toolbars()

        # Get internal and external toolbars
        internal_toolbars = get_class_values(ApplicationToolbars)
        external_toolbars = [
            toolbar_id
            for toolbar_id in app_toolbars.keys()
            if toolbar_id not in internal_toolbars
        ]

        # Default order for internal toolbars
        internal_toolbars_order = [
            ApplicationToolbars.File,
            ApplicationToolbars.Run,
            ApplicationToolbars.Debug,
            ApplicationToolbars.Profile,
            ApplicationToolbars.Main,
        ]

        # Check we didn't leave out any internal toolbar from the order above
        if DEV:
            if (
                (set(internal_toolbars) - set(internal_toolbars_order))
                != {ApplicationToolbars.WorkingDirectory}
            ):
                missing_toolbars = (
                    set(internal_toolbars)
                    - set(internal_toolbars_order)
                    - {ApplicationToolbars.WorkingDirectory}
                )

                raise SpyderAPIError(
                    f"The internal toolbar(s) {missing_toolbars} are not "
                    f"listed in the ordering of toolbars that is set below. "
                    f"Please add them to fix this error"
                )

        # Reorganize toolbars only if this is the first time Spyder starts or
        # new toolbars were added
        last_toolbars = self.get_conf("last_toolbars")
        if (
            not last_toolbars
            or set(last_toolbars) != set(app_toolbars.keys())
        ):
            logger.debug("Reorganize application toolbars")

            # We need to remove all toolbars first to organize them in the way
            # we want
            for toolbar in self._toolbarslist:
                self._plugin.main.removeToolBar(toolbar)

            # Add toolbars with the working directory to the right because it's
            # not clear where it ends, so users can have a hard time finding a
            # new toolbar in the interface if it's placed next to it.
            toolbars_order = internal_toolbars_order + external_toolbars
            for toolbar_id in (
                toolbars_order
                + [ApplicationToolbars.WorkingDirectory]
            ):
                toolbar = app_toolbars[toolbar_id]
                self._plugin.main.addToolBar(toolbar)
                toolbar.render()
        else:
            logger.debug("Render application toolbars")

            for toolbar in self._toolbarslist:
                toolbar.render()

    def save_last_toolbars(self):
        """Save the last available toolbars when the app is closed."""
        logger.debug("Saving current application toolbars")

        toolbars = []
        for toolbar in self._toolbarslist:
            toolbars.append(toolbar.objectName())

        self.set_conf('last_toolbars', toolbars)

    def save_last_visible_toolbars(self):
        """Save the last visible toolbars in our preferences."""
        if self.get_conf("toolbars_visible"):
            self._set_visible_toolbars()
        self._save_visible_toolbars()

    def load_last_visible_toolbars(self):
        """Load the last visible toolbars saved in our config system."""
        toolbars_names = self.get_conf('last_visible_toolbars')
        toolbars_visible = self.get_conf("toolbars_visible")

        # This is necessary to discard toolbars that were available in the last
        # session but are not on this one.
        visible_toolbars = []
        for name in toolbars_names:
            if name in self._APPLICATION_TOOLBARS:
                visible_toolbars.append(self._APPLICATION_TOOLBARS[name])

        # Update visible toolbars
        self._visible_toolbars = visible_toolbars

        # Show visible/hidden toolbars
        for toolbar in self._toolbarslist:
            if toolbar in self._visible_toolbars:
                toolbar.setVisible(toolbars_visible)
                toolbar.toggleViewAction().setChecked(toolbars_visible)
            else:
                toolbar.setVisible(False)
                toolbar.toggleViewAction().setChecked(False)

        self.update_actions()

    def create_toolbars_menu(self):
        """
        Populate the toolbars menu inside the view application menu.
        """
        main_section = ToolbarsMenuSections.Main
        secondary_section = ToolbarsMenuSections.Secondary
        default_toolbars = get_class_values(ApplicationToolbars)

        for toolbar_id, toolbar in self._ADDED_TOOLBARS.items():
            if toolbar:
                action = toolbar.toggleViewAction()

                # This is necessary to show the same visible toolbars both in
                # MainWindow and EditorMainWindow.
                action.triggered.connect(self.save_last_visible_toolbars)

                if not (PYSIDE2 or PYSIDE6):
                    # Modifying __class__ of a QObject created by C++ [1] seems
                    # to invalidate the corresponding Python object when PySide
                    # is used (changing __class__ of a QObject created in
                    # Python seems to work).
                    #
                    # [1] There are Qt functions such as
                    # QToolBar.toggleViewAction(), QToolBar.addAction(QString)
                    # and QMainWindow.addToolbar(QString), which return a
                    # pointer to an already existing QObject.
                    action.__class__ = QActionID

                # Register action
                id_ = f'toggle_view_{toolbar_id}'
                action.action_id = id_

                ACTION_REGISTRY.register_reference(
                    action,
                    id_,
                    self._plugin.NAME
                )

                # Add action to menu
                section = (
                    main_section
                    if toolbar_id in default_toolbars
                    else secondary_section
                )

                self.add_item_to_menu(
                    action,
                    menu=self.toolbars_menu,
                    section=section,
                )



# -----------------------------------------------------------------------------
# Copyright (c) 2024- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Helper classes to get and set keyboard shortcuts in Spyder.
"""

from __future__ import annotations

# Standard library imports
import functools
from collections.abc import Callable

# Third-party imports
from qtpy.QtCore import Qt
from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import QShortcut, QWidget

# Local imports
from spyder.api.config.mixins import SpyderConfigurationObserver
from spyder.config.manager import CONF
from spyder.plugins.shortcuts.utils import (
    ShortcutData,
    SHORTCUTS_FOR_WIDGETS_DATA,
)


class SpyderShortcutsMixin(SpyderConfigurationObserver):
    """Provide methods to get, set and register shortcuts for widgets."""

    def __init__(self) -> None:
        """Helper to get, set and register shortcuts for widgets."""
        super().__init__()

        # This is used to keep track of the widget shortcuts
        self._shortcuts: dict[(str, str), QShortcut] = {}

    def get_shortcut(
        self,
        name: str,
        context: str | None = None,
        plugin_name: str | None = None,
    ) -> str:
        """
        Get a shortcut sequence stored under the given name and context.

        Parameters
        ----------
        name: str
            The shortcut name (e.g. ``"run cell"``).
        context: str | None, optional
            Name of the shortcut context, e.g. ``"editor"`` for shortcuts
            that have effect when the :guilabel:`Editor` is focused or
            ``"_"`` for global shortcuts. If not set, the widget's
            :attr:`~spyder.api.plugins.SpyderPluginV2.CONF_SECTION` will
            be used as context.
        plugin_name: str | None, optional
            Name of the plugin where the shortcut is defined. Defaults to the
            context name; necessary for third-party plugins that have shortcuts
            with a context different from the plugin name.

        Returns
        -------
        shortcut: str
            Key sequence of the shortcut, e.g. ``"Ctrl+Enter"``.

        Raises
        ------
        configparser.NoOptionError
            If the shortcut does not exist in the configuration.
        """
        context = self.CONF_SECTION if context is None else context
        return CONF.get_shortcut(context, name, plugin_name)

    def set_shortcut(
        self,
        shortcut: str,
        name: str,
        context: str | None = None,
        plugin_name: str | None = None,
    ) -> None:
        """
        Set a shortcut sequence with a given name and context.

        Parameters
        ----------
        shortcut: str
            Key sequence of the shortcut, e.g. ``"Ctrl+Enter"``.
        name: str
            The shortcut name (e.g. ``"run cell"``).
        context: str | None, optional
            Name of the shortcut context, e.g. ``"editor"`` for shortcuts
            that have effect when the :guilabel:`Editor` is focused or
            ``"_"`` for global shortcuts. If not set, the widget's
            :attr:`~spyder.api.plugins.SpyderPluginV2.CONF_SECTION` will
            be used as context.
        plugin_name: str | None, optional
            Name of the plugin where the shortcut is defined. Defaults to the
            context name; necessary for third-party plugins that have shortcuts
            with a context different from the plugin name.

        Returns
        -------
        None

        Raises
        ------
        configparser.NoOptionError
            If the shortcut does not exist in the configuration.
        """
        context = self.CONF_SECTION if context is None else context
        CONF.set_shortcut(context, name, shortcut, plugin_name)

    def register_shortcut_for_widget(
        self,
        name: str,
        triggered: Callable,
        widget: QWidget | None = None,
        context: str | None = None,
        plugin_name: str | None = None,
    ) -> None:
        """
        Register a shortcut for a widget that inherits this mixin.

        Parameters
        ----------
        name: str
            The shortcut name (e.g. ``"run cell"``).
        triggered: Callable
            Callable (i.e. function or method) to be triggered by the shortcut.
        widget: QWidget, optional
            Widget to which this shortcut will be registered. If not set, the
            widget that calls this method will be used.
        context: str | None, optional
            Name of the shortcut context, e.g. ``"editor"`` for shortcuts
            that have effect when the :guilabel:`Editor` is focused or
            ``"_"`` for global shortcuts. If not set, the widget's
            :attr:`~spyder.api.plugins.SpyderPluginV2.CONF_SECTION` will
            be used as context.
        plugin_name: str | None, optional
            Name of the plugin where the shortcut is defined. Defaults to the
            context name; necessary for third-party plugins that have shortcuts
            with a context different from the plugin name.

        Returns
        -------
        None
        """
        context = self.CONF_SECTION if context is None else context
        widget = self if widget is None else widget

        # Name and context are saved in lowercase in our config system, so we
        # need to use them like that here.
        # Note: That's how the Python ConfigParser class saves options.
        name = name.lower()
        context = context.lower()

        # Register shortcurt for widget
        keystr = self.get_shortcut(name, context, plugin_name)
        self._register_shortcut(
            keystr, name, triggered, context, widget, plugin_name
        )

        # Add observer for shortcut so that it's updated when changed by users
        # in Preferences
        config_observer = functools.partial(
            self._register_shortcut,
            name=name,
            triggered=triggered,
            context=context,
            widget=widget,
            plugin_name=plugin_name,
        )

        self.add_configuration_observer(
            config_observer, option=f"{context}/{name}", section="shortcuts"
        )

        # Keep track of all widget shortcuts. This is necessary to show them in
        # Preferences.
        data = ShortcutData(
            qobject=None, name=name, context=context, plugin_name=plugin_name
        )
        if data not in SHORTCUTS_FOR_WIDGETS_DATA:
            SHORTCUTS_FOR_WIDGETS_DATA.append(data)

    def _register_shortcut(
        self,
        keystr: str,
        name: str,
        triggered: Callable,
        context: str,
        widget: QWidget,
        plugin_name: str | None,
    ) -> None:
        """
        Auxiliary function to register a shortcut for a widget.

        Parameters
        ----------
        keystr: str
            Key string for the shortcut, e.g. ``"Ctrl+Enter"``.
        name: str
            The shortcut name (e.g. ``"run cell"``).
        triggered: Callable
            Callable (i.e. function or method) to be triggered by the shortcut.
        context: str
            Name of the shortcut context, e.g. ``"editor"`` for shortcuts
            that have effect when the :guilabel:`Editor` is focused or
            ``"_"`` for global shortcuts.
        widget: QWidget
            Widget to which this shortcut will be registered.
        plugin_name: str | None
            Name of the plugin where the shortcut is defined.

        Returns
        -------
        None
        """
        # Disable current shortcut, if available
        current_shortcut = self._shortcuts.get((context, name, plugin_name))
        if current_shortcut:
            # Don't do the rest if we're trying to register the same shortcut
            # again. This happens at startup because shortcuts are registered
            # on widget creation and then the observer attached to the shortcut
            # tries to do it again after CONF.notify_all_observers() is called.
            if current_shortcut.key().toString() == keystr:
                return

            # Disable current shortcut to create a new one below
            current_shortcut.setEnabled(False)
            current_shortcut.deleteLater()
            self._shortcuts.pop((context, name, plugin_name))

        # Create a new shortcut
        new_shortcut = QShortcut(QKeySequence(keystr), widget)
        new_shortcut.activated.connect(triggered)
        new_shortcut.setContext(Qt.WidgetWithChildrenShortcut)

        # Save shortcut
        self._shortcuts[(context, name, plugin_name)] = new_shortcut



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Mixin to manage editing with multiple cursors.
"""

# Standard library imports
import functools
import itertools

# Third party imports
from qtpy.QtCore import Qt, QTimer, Slot
from qtpy.QtGui import (
    QColor, QFontMetrics, QPaintEvent, QPainter, QTextCursor, QKeyEvent
)
from qtpy.QtWidgets import QApplication, QMessageBox

# Local imports
from spyder.api.translations import _
from spyder.plugins.editor.api.decoration import TextDecoration
from spyder.utils.palette import SpyderPalette


class MultiCursorMixin:
    """Mixin to manage editing with multiple cursors."""

    def init_multi_cursor(self):
        """Initialize attrs and callbacks for multi-cursor functionality"""
        # actual default comes from setup_editor default args
        self.multi_cursor_enabled = False
        self.cursor_width = self.get_conf('cursor/width', section='main')
        self.overwrite_mode = self.overwriteMode()

        # Track overwrite manually when for painting reasons with multi-cursor
        self.setOverwriteMode(False)
        self.setCursorWidth(0)  # draw our own cursor
        self.extra_cursors = []
        self.cursor_blink_state = False
        self.cursor_blink_timer = QTimer(self)
        self.cursor_blink_timer.setInterval(
            QApplication.cursorFlashTime() // 2
        )
        self.cursor_blink_timer.timeout.connect(
            self._on_cursor_blinktimer_timeout
        )
        self.focus_in.connect(self.start_cursor_blink)
        self.focus_changed.connect(self.stop_cursor_blink)
        self.painted.connect(self.paint_cursors)
        self.multi_cursor_ignore_history = False
        self._drag_cursor = None

    def toggle_multi_cursor(self, enabled):
        """Enable/disable multi-cursor editing."""
        self.multi_cursor_enabled = enabled

        # TODO: Any restrictions on enabling? only python-like? only code?
        if not enabled:
            self.clear_extra_cursors()

    def add_column_cursor(self, event):
        """
        Add a cursor on each row between primary cursor and click location.
        """

        if not self.multi_cursor_enabled:
            return

        self.multi_cursor_ignore_history = True

        # Ctrl-Shift-Alt click adds colum of cursors towards primary
        # cursor
        cursor_for_pos = self.cursorForPosition(event.pos())
        first_cursor = self.textCursor()
        anchor_block = first_cursor.block()
        anchor_col = first_cursor.anchor() - anchor_block.position()
        pos_block = cursor_for_pos.block()
        pos_col = cursor_for_pos.positionInBlock()

        # Warn user of potentially large number of cursors
        if abs(anchor_block.blockNumber() - pos_block.blockNumber()) >= 1000:
            response = QMessageBox.warning(
                self,
                _("Large number of cursors!"),
                _(
                    "A large numbers of text cursors can cause Spyder to "
                    "become unresponsive. Do you want to continue?"
                ),
                QMessageBox.Ok | QMessageBox.Cancel
            )
            if not response == QMessageBox.Ok:
                self.multi_cursor_ignore_history = False
                return

        # Move primary cursor to pos_col
        p_col = min(len(anchor_block.text()), pos_col)

        # block.length() includes line separator? just \n?
        # use len(block.text()) instead
        first_cursor.setPosition(anchor_block.position() + p_col,
                                 QTextCursor.MoveMode.KeepAnchor)
        self.setTextCursor(first_cursor)
        block = anchor_block
        while block != pos_block:
            # Get the next block
            if anchor_block < pos_block:
                block = block.next()
            else:
                block = block.previous()

            # Add a cursor for this block
            if block.isVisible() and block.isValid():
                cursor = QTextCursor(first_cursor)

                a_col = min(len(block.text()), anchor_col)
                cursor.setPosition(block.position() + a_col,
                                   QTextCursor.MoveMode.MoveAnchor)
                p_col = min(len(block.text()), pos_col)
                cursor.setPosition(block.position() + p_col,
                                   QTextCursor.MoveMode.KeepAnchor)
                self.extra_cursors.append(cursor)

        self.merge_extra_cursors(True)
        self.multi_cursor_ignore_history = False
        self.cursorPositionChanged.emit()

    def add_remove_cursor(self, event):
        """Add or remove extra cursor on mouse click event"""

        if not self.multi_cursor_enabled:
            return

        self.multi_cursor_ignore_history = True

        # Move existing primary cursor to extra_cursors list and set
        # new primary cursor
        cursor_for_pos = self.cursorForPosition(event.pos())
        old_cursor = self.textCursor()

        # Don't attempt to remove cursor if there's only one
        removed_cursor = False
        if self.extra_cursors:
            same_cursor = None
            for cursor in self.all_cursors:
                if cursor_for_pos.position() == cursor.position():
                    same_cursor = cursor
                    break
            if same_cursor is not None:
                removed_cursor = True
                if same_cursor in self.extra_cursors:
                    # cursor to be removed was not primary
                    self.extra_cursors.remove(same_cursor)
                else:
                    # cursor to be removed is primary cursor
                    # pick a new primary by position
                    new_primary = max(
                        self.extra_cursors,
                        key=lambda cursor: cursor.position()
                    )
                    self.extra_cursors.remove(new_primary)
                    self.setTextCursor(new_primary)

                # Possibly clear selection of removed cursor
                self.set_extra_cursor_selections()

        if not removed_cursor:
            self.setTextCursor(cursor_for_pos)
            self.extra_cursors.append(old_cursor)
            self.merge_extra_cursors(True)

        self.multi_cursor_ignore_history = False
        self.cursorPositionChanged.emit()

    def add_cursor_up(self):
        if self.multi_cursor_enabled:
            self.extra_cursors.append(self.textCursor())
            self.moveCursor(QTextCursor.MoveOperation.Up)
            self.merge_extra_cursors(True)

    def add_cursor_down(self):
        if self.multi_cursor_enabled:
            self.extra_cursors.append(self.textCursor())
            self.moveCursor(QTextCursor.MoveOperation.Down)
            self.merge_extra_cursors(True)

    def set_extra_cursor_selections(self):
        selections = []
        for cursor in self.extra_cursors:
            extra_selection = TextDecoration(
                cursor, draw_order=5, kind="extra_cursor_selection"
            )

            extra_selection.set_foreground(
                QColor(SpyderPalette.COLOR_TEXT_1)
            )
            extra_selection.set_background(
                QColor(SpyderPalette.COLOR_ACCENT_2)
            )

            selections.append(extra_selection)
        self.set_extra_selections('extra_cursor_selections', selections)

    def clear_extra_cursors(self):
        """Remove all extra cursors"""
        self.extra_cursors = []
        self.set_extra_selections('extra_cursor_selections', [])

    @property
    def all_cursors(self):
        """Return list of all extra_cursors (if any) plus the primary cursor"""
        return self.extra_cursors + [self.textCursor()]

    def merge_extra_cursors(self, increasing_position):
        """Merge overlapping cursors"""
        if not self.extra_cursors:
            return

        previous_history = self.multi_cursor_ignore_history
        self.multi_cursor_ignore_history = True

        while True:
            cursor_was_removed = False

            cursors = self.all_cursors
            main_cursor = cursors[-1]
            cursors.sort(key=lambda cursor: cursor.position())

            for i, cursor1 in enumerate(cursors[:-1]):
                if cursor_was_removed:
                    break  # list will be modified, so re-start at while loop

                for cursor2 in cursors[i + 1:]:
                    # given cursors.sort, pos1 should be <= pos2
                    pos1 = cursor1.position()
                    pos2 = cursor2.position()
                    anchor1 = cursor1.anchor()
                    anchor2 = cursor2.anchor()

                    if not pos1 == pos2:
                        continue  # only merge coincident cursors

                    if cursor1 is main_cursor:
                        # swap cursors to keep main_cursor
                        cursor1, cursor2 = cursor2, cursor1

                    self.extra_cursors.remove(cursor1)
                    cursor_was_removed = True

                    # reposition cursor we're keeping
                    positions = sorted([pos1, anchor1, anchor2])
                    if not increasing_position:
                        positions.reverse()
                    cursor2.setPosition(
                        positions[0],
                        QTextCursor.MoveMode.MoveAnchor
                    )
                    cursor2.setPosition(
                        positions[2],
                        QTextCursor.MoveMode.KeepAnchor
                    )
                    if cursor2 is main_cursor:
                        self.setTextCursor(cursor2)
                    break

            if not cursor_was_removed:
                break

        self.set_extra_cursor_selections()
        self.multi_cursor_ignore_history = previous_history

    @Slot(QKeyEvent)
    def handle_multi_cursor_keypress(self, event: QKeyEvent):
        """Re-Implement keyEvent handler for multi-cursor"""

        key = event.key()
        ctrl = event.modifiers() & Qt.KeyboardModifier.ControlModifier
        alt = event.modifiers() & Qt.KeyboardModifier.AltModifier
        shift = event.modifiers() & Qt.KeyboardModifier.ShiftModifier

        # ---- Handle insert
        if key == Qt.Key.Key_Insert and not (ctrl or alt or shift):
            self.overwrite_mode = not self.overwrite_mode
            return

        self.textCursor().beginEditBlock()
        self.multi_cursor_ignore_history = True

        # Handle all signals before editing text
        cursors = []
        accepted = []
        for cursor in self.all_cursors:
            self.setTextCursor(cursor)
            event.ignore()
            self.sig_key_pressed.emit(event)
            cursors.append(self.textCursor())
            accepted.append(event.isAccepted())

        increasing_position = True
        new_cursors = []
        for skip, cursor in zip(accepted, cursors):
            self.setTextCursor(cursor)
            if skip:
                # Text folding swallows most input to prevent typing on folded
                # lines.
                pass
            # ---- Handle Tab
            elif key == Qt.Key.Key_Tab and not ctrl:  # ctrl-tab is shortcut
                # Don't do intelligent tab with multi-cursor to skip
                # calls to do_completion. Avoiding completions with multi
                # cursor is much easier than solving all the edge cases.
                self.indent(force=self.tab_mode)
            elif key == Qt.Key.Key_Backtab and not ctrl:
                increasing_position = False
                # TODO: Ignore indent level of neighboring lines and simply
                # indent by 1 level at a time. Cursor update order can
                # make this unpredictable otherwise.
                self.unindent(force=self.tab_mode)
            # ---- Handle enter/return
            elif key in (Qt.Key_Enter, Qt.Key_Return):
                if not shift and not ctrl:
                    if (
                        self.add_colons_enabled and
                        self.is_python_like() and
                        self.autoinsert_colons()
                    ):
                        self.insert_text(':' + self.get_line_separator())
                        if self.strip_trailing_spaces_on_modify:
                            self.fix_and_strip_indent()
                        else:
                            self.fix_indent()
                    else:
                        cur_indent = self.get_block_indentation(
                            self.textCursor().blockNumber())
                        self._handle_keypress_event(event)

                        # Check if we're in a comment or a string at the
                        # current position
                        cmt_or_str_cursor = self.in_comment_or_string()

                        # Check if the line start with a comment or string
                        cursor = self.textCursor()
                        cursor.setPosition(
                            cursor.block().position(),
                            QTextCursor.KeepAnchor
                        )
                        cmt_or_str_line_begin = self.in_comment_or_string(
                            cursor=cursor
                        )

                        # Check if we are in a comment or a string
                        cmt_or_str = (
                            cmt_or_str_cursor and cmt_or_str_line_begin
                        )

                        if self.strip_trailing_spaces_on_modify:
                            self.fix_and_strip_indent(
                                comment_or_string=cmt_or_str,
                                cur_indent=cur_indent
                            )
                        else:
                            self.fix_indent(
                                comment_or_string=cmt_or_str,
                                cur_indent=cur_indent
                            )
            # ---- Intelligent backspace handling
            elif key == Qt.Key_Backspace and not shift and not ctrl:
                increasing_position = False
                if self.has_selected_text() or not self.intelligent_backspace:
                    self._handle_keypress_event(event)
                else:
                    leading_text = self.get_text('sol', 'cursor')
                    leading_length = len(leading_text)
                    trailing_spaces = (
                        leading_length - len(leading_text.rstrip())
                    )
                    trailing_text = self.get_text('cursor', 'eol')
                    matches = ('()', '[]', '{}', '\'\'', '""')
                    if (
                        not leading_text.strip() and
                        (leading_length > len(self.indent_chars))
                    ):
                        if leading_length % len(self.indent_chars) == 0:
                            self.unindent()
                        else:
                            self._handle_keypress_event(event)
                    elif trailing_spaces and not trailing_text.strip():
                        self.remove_suffix(leading_text[-trailing_spaces:])
                    elif (
                        leading_text and
                        trailing_text and
                        (leading_text[-1] + trailing_text[0] in matches)
                    ):
                        cursor = self.textCursor()
                        cursor.movePosition(QTextCursor.PreviousCharacter)
                        cursor.movePosition(
                            QTextCursor.NextCharacter,
                            QTextCursor.KeepAnchor, 2
                        )
                        cursor.removeSelectedText()
                    else:
                        self._handle_keypress_event(event)
            # ---- Handle home, end
            elif key == Qt.Key.Key_Home:
                increasing_position = False
                self.stdkey_home(shift, ctrl)
            elif key == Qt.Key.Key_End:
                # See spyder-ide/spyder#495: on MacOS X, it is necessary to
                # redefine this basic action which should have been implemented
                # natively
                self.stdkey_end(shift, ctrl)
            # ---- Use default handler for cursor (text)
            else:
                if key in (Qt.Key.Key_Up, Qt.Key.Key_Left):
                    increasing_position = False
                if (
                    key in (Qt.Key.Key_Up, Qt.Key.Key_Down)
                    and cursor.verticalMovementX() == -1
                ):
                    # Builtin handler somehow does not set verticalMovementX
                    # when moving up and down (but works fine for single
                    # cursor somehow)
                    # TODO: Why? Are we forgetting something?
                    x = self.cursorRect(cursor).x()
                    cursor.setVerticalMovementX(x)
                    self.setTextCursor(cursor)
                self._handle_keypress_event(event)

            # Update edited extra_cursors
            new_cursors.append(self.textCursor())

        self.extra_cursors = new_cursors[:-1]
        self.merge_extra_cursors(increasing_position)
        self.textCursor().endEditBlock()
        self.multi_cursor_ignore_history = False
        self.cursorPositionChanged.emit()
        event.accept()  # TODO when to pass along keypress or not

    def _on_cursor_blinktimer_timeout(self):
        """
        Text cursor blink timer generates paint events and inverts draw state
        """
        self.cursor_blink_state = not self.cursor_blink_state
        if self.isVisible():
            self.viewport().update()

    @Slot(QPaintEvent)
    def paint_cursors(self, event):
        """Paint all cursors"""
        if self.overwrite_mode:
            font = self.font()
            cursor_width = QFontMetrics(font).horizontalAdvance(" ")
        else:
            cursor_width = self.cursor_width

        qp = QPainter()
        qp.begin(self.viewport())
        offset = self.contentOffset()
        content_offset_y = offset.y()
        qp.setBrushOrigin(offset)
        editable = not self.isReadOnly()
        flags = (
            self.textInteractionFlags()
            & Qt.TextInteractionFlag.TextSelectableByKeyboard
        )

        if self._drag_cursor is not None and (editable or flags):
            cursor = self._drag_cursor
            block = cursor.block()
            if block.isVisible():
                block_top = int(self.blockBoundingGeometry(block).top())
                offset.setY(block_top + content_offset_y)
                layout = block.layout()
                if layout is not None:  # Fix exceptions in test_flag_painting
                    layout.drawCursor(
                        qp,
                        offset,
                        cursor.positionInBlock(),
                        cursor_width
                    )

        draw_cursor = self.cursor_blink_state and (editable or flags)

        for cursor in self.all_cursors:
            block = cursor.block()
            if draw_cursor and block.isVisible():
                block_top = int(self.blockBoundingGeometry(block).top())
                offset.setY(block_top + content_offset_y)
                layout = block.layout()
                if layout is not None:
                    layout.drawCursor(
                        qp,
                        offset,
                        cursor.positionInBlock(),
                        cursor_width
                    )
        qp.end()

    @Slot()
    def start_cursor_blink(self):
        """Start manually updating the cursor(s) blink state: Show cursors."""
        self.cursor_blink_state = True
        self.cursor_blink_timer.start()

    @Slot()
    def stop_cursor_blink(self):
        """Stop manually updating the cursor(s) blink state: Hide cursors."""
        self.cursor_blink_state = False
        self.cursor_blink_timer.stop()

    def multi_cursor_copy(self):
        """
        Join all cursor selections in position sorted order by line_separator,
        and put text to clipboard.
        """
        cursors = self.all_cursors
        cursors.sort(key=lambda cursor: cursor.position())
        selections = []
        for cursor in cursors:
            text = cursor.selectedText().replace(
                "\u2029",
                self.get_line_separator()
            )
            selections.append(text)
        clip_text = self.get_line_separator().join(selections)
        QApplication.clipboard().setText(clip_text)

    def multi_cursor_cut(self):
        """Multi-cursor copy then removeSelectedText"""
        self.multi_cursor_copy()
        self.textCursor().beginEditBlock()
        for cursor in self.all_cursors:
            cursor.removeSelectedText()

        # Merge direction doesn't matter here as all selections are removed
        self.merge_extra_cursors(True)
        self.textCursor().endEditBlock()

    def multi_cursor_paste(self, clip_text):
        """
        Paste one line per cursor or entire clipboard per cursor depending on
        settings.
        """
        main_cursor = self.textCursor()
        main_cursor.beginEditBlock()
        cursors = self.all_cursors
        cursors.sort(key=lambda cursor: cursor.position())
        self.skip_rstrip = True
        self.sig_will_paste_text.emit(clip_text)
        lines = clip_text.splitlines()
        
        if self.get_conf('multicursor_paste/always_full'):
            lines = itertools.repeat(clip_text)
        elif self.get_conf('multicursor_paste/conditional_spread'):
            if len(lines) != len(cursors):
                lines = itertools.repeat(clip_text)
        elif self.get_conf('multicursor_paste/always_spread'):
            if len(lines) == 1:
                lines = itertools.repeat(lines[0])

        self.multi_cursor_ignore_history = True
        for cursor, text in zip(cursors, lines):
            self.setTextCursor(cursor)
            cursor.insertText(text)
            # handle extra lines or extra cursors?

        self.setTextCursor(main_cursor)
        self.multi_cursor_ignore_history = False
        self.cursorPositionChanged.emit()

        # Merge direction doesn't matter here as all selections are removed
        self.merge_extra_cursors(True)
        main_cursor.endEditBlock()
        self.sig_text_was_inserted.emit()
        self.skip_rstrip = False

    def for_each_cursor(self, method, merge_increasing=True):
        """Wrap callable to execute once for each cursor"""
        @functools.wraps(method)
        def wrapper():
            self.textCursor().beginEditBlock()
            new_cursors = []
            self.multi_cursor_ignore_history = True
            for cursor in self.all_cursors:
                self.setTextCursor(cursor)

                # May call setTtextCursor with modified copy
                method()

                # Get modified cursor to re-add to extra_cursors
                new_cursors.append(self.textCursor())

            # re-add extra cursors
            self.clear_extra_cursors()
            if self.multi_cursor_enabled:
                self.extra_cursors = new_cursors[:-1]
            self.setTextCursor(new_cursors[-1])
            self.merge_extra_cursors(merge_increasing)
            self.textCursor().endEditBlock()
            self.multi_cursor_ignore_history = False
            self.cursorPositionChanged.emit()

        return wrapper

    def clears_extra_cursors(self, method):
        """Wrap callable to clear extra_cursors prior to calling"""
        @functools.wraps(method)
        def wrapper():
            self.clear_extra_cursors()
            method()

        return wrapper

    def restrict_single_cursor(self, method):
        """Wrap callable to only execute if there is a single cursor"""
        @functools.wraps(method)
        def wrapper():
            if not self.extra_cursors:
                method()

        return wrapper

    def go_to_next_cell(self):
        """
        Reimplement TextEditBaseWidget.go_to_next_cell to clear extra cursors.
        """
        self.clear_extra_cursors()
        super().go_to_next_cell()

    def go_to_previous_cell(self):
        """
        Reimplement TextEditBaseWidget.go_to_previous_cell to clear extra
        cursors.
        """
        self.clear_extra_cursors()
        super().go_to_previous_cell()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------


"""
Widgets for the IPython Console.
"""

from .control import ControlWidget, PageControlWidget
from .debugging import DebuggingWidget
from .help import HelpWidget
from .namespacebrowser import NamepaceBrowserWidget
from .figurebrowser import FigureBrowserWidget
from .kernelconnect import KernelConnectionDialog
from .restartdialog import ConsoleRestartDialog
from .status import MatplotlibStatus, PythonEnvironmentStatus

# ShellWidget contains the other widgets and ClientWidget
# contains it
from .shell import ShellWidget
from .client import ClientWidget

# Constant for different types of completion available
COMPLETION_WIDGET_TYPE = {0: "droplist", 1: "ncurses", 2: "plain"}



# -----------------------------------------------------------------------------
# Copyright (c) 2017- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Deprecated aliases of classes in :mod:`spyder.plugins.editor.api`.

.. deprecated:: 6.2

    This module will issue a :exc:`DeprecationWarning` in Spyder 6.2 and
    be removed in Spyder 7.0, as with the migration of the :guilabel:`Editor`
    to the new plugin API these classes have been canonically accessed
    directly from the Editor plugin since Spyder 6.0.

    Use :class:`spyder.plugins.editor.api.editorextension.EditorExtension`
    and :class:`spyder.plugins.editor.api.panel.Panel` instead.
"""

from spyder.plugins.editor.api.editorextension import EditorExtension
from spyder.plugins.editor.api.panel import Panel


__all__ = ["EditorExtension", "Panel"]



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Customized combobox widgets."""

# pylint: disable=C0103
# pylint: disable=R0903
# pylint: disable=R0911
# pylint: disable=R0201


# Standard library imports
import glob
import os
import os.path as osp

# Third party imports
from qtpy import PYSIDE2
from qtpy.QtCore import QEvent, Qt, QTimer, QUrl, Signal, QSize
from qtpy.QtGui import QFont
from qtpy.QtWidgets import (
    QComboBox, QCompleter, QLineEdit, QSizePolicy, QToolTip)

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.comboboxes import SpyderComboBox
from spyder.utils.stylesheet import APP_STYLESHEET
from spyder.widgets.helperwidgets import ClearLineEdit, IconLineEdit


class BaseComboBox(SpyderComboBox):
    """Editable combo box base class"""
    valid = Signal(bool, bool)
    sig_tab_pressed = Signal(bool)

    sig_resized = Signal(QSize, QSize)
    """
    This signal is emitted to inform the widget has been resized.

    Parameters
    ----------
    size: QSize
        The new size of the widget.
    old_size: QSize
        The previous size of the widget.
    """

    def __init__(self, parent, items_elide_mode=None):
        super().__init__(parent, items_elide_mode)
        self.setEditable(True)
        self.setCompleter(QCompleter(self))
        self.selected_text = self.currentText()

    # --- Qt overrides
    def event(self, event):
        """Qt Override.

        Filter tab keys and process double tab keys.
        """

        # Type check: Prevent error in PySide where 'event' may be of type
        # QtGui.QPainter (for whatever reason).
        if not isinstance(event, QEvent):
            return True

        if (event.type() == QEvent.KeyPress) and (event.key() == Qt.Key_Tab):
            self.sig_tab_pressed.emit(True)
            return True
        return super().event(event)
    
    def focusOutEvent(self, event):
        """
        Qt Override.

        Handle focus out event to prevent changing current text with some other
        entry in the history that could match the current text in a case
        insensitive manner.
        See spyder-ide/spyder#23597
        """
        self.add_current_text_if_valid()
        super().focusOutEvent(event)

    def keyPressEvent(self, event):
        """Qt Override.

        Handle key press events.
        """
        if event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            if self.add_current_text_if_valid():
                self.selected()
                self.hide_completer()
        elif event.key() == Qt.Key_Escape:
            self.set_current_text(self.selected_text)
            self.hide_completer()
        else:
            super().keyPressEvent(event)

    def resizeEvent(self, event):
        """
        Emit a resize signal for widgets that need to adapt its size.
        """
        super().resizeEvent(event)
        self.sig_resized.emit(event.size(), event.oldSize())

    # --- Own methods
    def is_valid(self, qstr):
        """
        Return True if string is valid
        Return None if validation can't be done
        """
        pass

    def selected(self):
        """Action to be executed when a valid item has been selected"""
        self.valid.emit(True, True)

    def add_text(self, text):
        """Add text to combo box: add a new item if text is not found in
        combo box items."""
        index = self.findText(text, Qt.MatchCaseSensitive)
        while index != -1:
            self.removeItem(index)
            index = self.findText(text, Qt.MatchCaseSensitive)
        self.insertItem(0, text)
        index = self.findText('')
        if index != -1:
            self.removeItem(index)
            self.insertItem(0, '')
            if text != '':
                self.setCurrentIndex(1)
            else:
                self.setCurrentIndex(0)
        else:
            self.setCurrentIndex(0)
        self.selected_text = text

    def set_current_text(self, text):
        """Sets the text of the QLineEdit of the QComboBox."""
        self.lineEdit().setText(str(text))

    def add_current_text(self):
        """Add current text to combo box history (convenient method)"""
        text = self.currentText()
        self.add_text(text)

    def add_current_text_if_valid(self):
        """Add current text to combo box history if valid"""
        valid = self.is_valid(self.currentText())
        if valid or valid is None:
            self.add_current_text()
            return True
        else:
            self.set_current_text(self.selected_text)

    def hide_completer(self):
        """Hides the completion widget."""
        self.setCompleter(QCompleter([], self))


class PatternComboBox(BaseComboBox):
    """Search pattern combo box"""

    def __init__(
        self,
        parent,
        items=None,
        tip=None,
        adjust_to_minimum=True,
        id_=None,
        items_elide_mode=None,
    ):
        if not PYSIDE2:
            super().__init__(parent, items_elide_mode)
        else:
            BaseComboBox.__init__(self, parent, items_elide_mode)

        if adjust_to_minimum:
            self.setSizeAdjustPolicy(
                QComboBox.AdjustToMinimumContentsLengthWithIcon
            )

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        if items is not None:
            self.addItems(items)
        if tip is not None:
            self.setToolTip(tip)
        if id_ is not None:
            self.ID = id_

        # Use a line edit with a clear button inside it.
        # Note: The method Qt offers for this (setClearButtonEnabled) adds a
        # button whose icon can't be easily stylized.
        self.setLineEdit(ClearLineEdit(self, reposition_button=True))


class EditableComboBox(BaseComboBox):
    """
    Editable combo box + Validate
    """

    def __init__(self, parent):
        if not PYSIDE2:
            super().__init__(parent)
        else:
            BaseComboBox.__init__(self, parent)

        self.font = QFont()
        self.selected_text = self.currentText()

        # Widget setup
        self.setSizeAdjustPolicy(
            QComboBox.AdjustToMinimumContentsLengthWithIcon
        )

        # Signals
        self.editTextChanged.connect(self.validate)
        self.tips = {True: _("Press enter to validate this entry"),
                     False: _('This entry is incorrect')}

    def show_tip(self, tip=""):
        """Show tip"""
        QToolTip.showText(self.mapToGlobal(self.pos()), tip, self)

    def selected(self):
        """Action to be executed when a valid item has been selected"""
        BaseComboBox.selected(self)
        self.selected_text = self.currentText()

    def validate(self, qstr, editing=True):
        """Validate entered path"""
        if self.selected_text == qstr and qstr != '':
            self.valid.emit(True, True)
            return

        valid = self.is_valid(qstr)
        if editing:
            if valid:
                self.valid.emit(True, False)
            else:
                self.valid.emit(False, False)


class PathComboBox(EditableComboBox):
    """
    QComboBox handling path locations
    """
    open_dir = Signal(str)

    def __init__(self, parent, adjust_to_contents=False, id_=None,
                 elide_text=False, ellipsis_place=Qt.ElideLeft):
        if not PYSIDE2:
            super().__init__(parent)
        else:
            EditableComboBox.__init__(self, parent)

        # Replace the default lineedit with a custom one with icon display
        # and elided text
        lineedit = IconLineEdit(self, elide_text=elide_text,
                                ellipsis_place=ellipsis_place)

        # Widget setup
        if adjust_to_contents:
            self.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        else:
            self.setSizeAdjustPolicy(
                QComboBox.AdjustToMinimumContentsLengthWithIcon
            )
            self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.tips = {True: _("Press enter to validate this path"),
                     False: ''}
        self.setLineEdit(lineedit)

        # Signals
        self.highlighted.connect(self.add_tooltip_to_highlighted_item)
        self.sig_tab_pressed.connect(self.tab_complete)
        self.valid.connect(lineedit.update_status)

        if id_ is not None:
            self.ID = id_

    # --- Qt overrides
    def focusInEvent(self, event):
        """Handle focus in event restoring to display the status icon."""
        show_status = getattr(self.lineEdit(), 'show_status_icon', None)
        if show_status:
            show_status()
        super().focusInEvent(event)

    def focusOutEvent(self, event):
        """Handle focus out event restoring the last valid selected path."""
        # Calling asynchronously the 'add_current_text' to avoid crash
        # https://groups.google.com/group/spyderlib/browse_thread/thread/2257abf530e210bd
        if not self.is_valid():
            lineedit = self.lineEdit()

            # Avoid error when lineedit is no longer available (probably
            # because this widget's parent was garbage collected).
            # Fixes spyder-ide/spyder#23361
            try:
                QTimer.singleShot(
                    50, lambda: lineedit.setText(self.selected_text)
                )
            except RuntimeError:
                pass

        hide_status = getattr(self.lineEdit(), 'hide_status_icon', None)
        if hide_status:
            hide_status()
        super().focusOutEvent(event)

    # --- Own methods
    def _complete_options(self):
        """Find available completion options."""
        text = str(self.currentText())
        opts = glob.glob(text + "*")
        opts = sorted([opt for opt in opts if osp.isdir(opt)])

        completer = QCompleter(opts, self)
        qss = str(APP_STYLESHEET)
        completer.popup().setStyleSheet(qss)
        self.setCompleter(completer)

        return opts

    def tab_complete(self):
        """
        If there is a single option available one tab completes the option.
        """
        opts = self._complete_options()
        if len(opts) == 0:
            return
        if len(opts) == 1:
            self.set_current_text(opts[0] + os.sep)
            self.hide_completer()
        else:
            self.set_current_text(osp.commonprefix(opts))
            self.completer().complete()

    def is_valid(self, qstr=None):
        """Return True if string is valid"""
        if qstr is None:
            qstr = self.currentText()
        return osp.isdir(str(qstr))

    def selected(self):
        """Action to be executed when a valid item has been selected"""
        self.selected_text = self.currentText()
        self.valid.emit(True, True)
        self.open_dir.emit(self.selected_text)

    def add_current_text(self):
        """
        Add current text to combo box history (convenient method).
        If path ends in os separator ("\" windows, "/" unix) remove it.
        """
        text = self.currentText()
        if osp.isdir(text) and text:
            if text[-1] == os.sep:
                text = text[:-1]
        self.add_text(text)

    def add_tooltip_to_highlighted_item(self, index):
        """
        Add a tooltip showing the full path of the currently highlighted item
        of the PathComboBox.
        """
        self.setItemData(index, self.itemText(index), Qt.ToolTipRole)


class UrlComboBox(PathComboBox):
    """
    QComboBox handling urls
    """
    def __init__(self, parent, adjust_to_contents=False, id_=None):
        if not PYSIDE2:
            super().__init__(parent, adjust_to_contents)
        else:
            PathComboBox.__init__(self, parent, adjust_to_contents)

        line_edit = QLineEdit(self)
        self.setLineEdit(line_edit)
        self.editTextChanged.disconnect(self.validate)

        if id_ is not None:
            self.ID = id_

    def is_valid(self, qstr=None):
        """Return True if string is valid"""
        if qstr is None:
            qstr = self.currentText()
        return QUrl(qstr).isValid()


class FileComboBox(PathComboBox):
    """
    QComboBox handling File paths
    """
    def __init__(self, parent=None, adjust_to_contents=False,
                 default_line_edit=False):
        if not PYSIDE2:
            super().__init__(parent, adjust_to_contents)
        else:
            PathComboBox.__init__(self, parent, adjust_to_contents)

        if default_line_edit:
            line_edit = QLineEdit(self)
            self.setLineEdit(line_edit)

        # Widget setup
        if adjust_to_contents:
            self.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        else:
            self.setSizeAdjustPolicy(QComboBox.AdjustToContentsOnFirstShow)
            self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

    def is_valid(self, qstr=None):
        """Return True if string is valid."""
        if qstr is None:
            qstr = self.currentText()
        valid = osp.isfile(str(qstr)) or osp.isdir(str(qstr))
        return valid

    def tab_complete(self):
        """
        If there is a single option available one tab completes the option.
        """
        opts = self._complete_options()
        if len(opts) == 1:
            text = opts[0]
            if osp.isdir(text):
                text = text + os.sep
            self.set_current_text(text)
            self.hide_completer()
        else:
            self.completer().complete()

    def _complete_options(self):
        """Find available completion options."""
        text = str(self.currentText())
        opts = glob.glob(text + "*")
        opts = sorted([opt for opt in opts
                       if osp.isdir(opt) or osp.isfile(opt)])

        completer = QCompleter(opts, self)
        qss = str(APP_STYLESHEET)
        completer.popup().setStyleSheet(qss)
        self.setCompleter(completer)

        return opts


def is_module_or_package(path):
    """Return True if path is a Python module/package"""
    is_module = osp.isfile(path) and osp.splitext(path)[1] in ('.py', '.pyw')
    is_package = osp.isdir(path) and osp.isfile(osp.join(path, '__init__.py'))
    return is_module or is_package


class PythonModulesComboBox(PathComboBox):
    """
    QComboBox handling Python modules or packages path
    (i.e. .py, .pyw files *and* directories containing __init__.py)
    """
    def __init__(self, parent, adjust_to_contents=False, id_=None):
        if not PYSIDE2:
            super().__init__(parent, adjust_to_contents)
        else:
            PathComboBox.__init__(self, parent, adjust_to_contents)

        if id_ is not None:
            self.ID = id_

    def is_valid(self, qstr=None):
        """Return True if string is valid"""
        if qstr is None:
            qstr = self.currentText()
        return is_module_or_package(str(qstr))

    def selected(self):
        """Action to be executed when a valid item has been selected"""
        EditableComboBox.selected(self)
        self.open_dir.emit(self.currentText())



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Spyder base configuration management

This file only deals with non-GUI configuration features
(in other words, we won't import any PyQt object here, avoiding any
sip API incompatibility issue in spyder's non-gui modules)
"""

from __future__ import annotations

from collections.abc import Callable
from glob import glob
import locale
import os
import os.path as osp
import re
import shutil
import sys
import tempfile
import uuid
import warnings

# Local imports
from spyder import __version__
from spyder.utils import encoding

#==============================================================================
# Only for development
#==============================================================================
# To activate/deactivate certain things for development
# SPYDER_DEV is (and *only* has to be) set in bootstrap.py
DEV = os.environ.get('SPYDER_DEV')

# Manually override whether the dev configuration directory is used.
USE_DEV_CONFIG_DIR = os.environ.get('SPYDER_USE_DEV_CONFIG_DIR')

# Get a random id for the safe-mode config dir
CLEAN_DIR_ID = str(uuid.uuid4()).split('-')[-1]


def get_safe_mode():
    """
    Make Spyder use a temp clean configuration directory for testing
    purposes SPYDER_SAFE_MODE can be set using the --safe-mode option.
    """
    return bool(os.environ.get('SPYDER_SAFE_MODE'))


def running_under_pytest():
    """
    Return True if currently running under pytest.

    This function is used to do some adjustment for testing. The environment
    variable SPYDER_PYTEST is defined in conftest.py.
    """
    return bool(os.environ.get('SPYDER_PYTEST'))


def running_remoteclient_tests():
    """
    Return True if currently running the remoteclient tests.

    This function is used to do some adjustment for testing. The environment
    variable SPYDER_TEST_REMOTE_CLIENT is 'true' in conftest.py.
    """
    return bool(os.environ.get("SPYDER_TEST_REMOTE_CLIENT") == "true")


def running_in_ci():
    """Return True if currently running under CI."""
    return bool(os.environ.get('CI'))


def running_in_ci_with_conda():
    """Return True if currently running under CI with conda packages."""
    return running_in_ci() and os.environ.get('USE_CONDA', None) == 'true'


def running_in_binder():
    """Return True if currently running in Binder."""
    return (
        bool(os.environ.get("BINDER_REPO_URL"))
        and "spyder-ide/binder-environments" in os.environ["BINDER_REPO_URL"]
    )


def building_autodoc():
    """Return True if executing under Sphinx Autodoc, False otherwise."""
    return bool(os.environ.get('SPHINX_AUTODOC'))


def is_stable_version(version):
    """
    Return true if version is stable, i.e. with letters in the final component.

    Stable version examples: ``1.2``, ``1.3.4``, ``1.0.5``.
    Non-stable version examples: ``1.3.4beta``, ``0.1.0rc1``, ``3.0.0dev0``.
    """
    if not isinstance(version, tuple):
        version = version.split('.')
    last_part = version[-1]

    if not re.search(r'[a-zA-Z]', last_part):
        return True
    else:
        return False


def use_dev_config_dir(use_dev_config_dir=USE_DEV_CONFIG_DIR):
    """Return whether the dev configuration directory should used."""
    if use_dev_config_dir is not None:
        if use_dev_config_dir.lower() in {'false', '0'}:
            use_dev_config_dir = False
    else:
        use_dev_config_dir = DEV or not is_stable_version(__version__)

    return use_dev_config_dir


#==============================================================================
# Debug helpers
#==============================================================================
# This is needed after restarting and using debug_print
STDOUT = sys.stdout
STDERR = sys.stderr


def get_debug_level():
    debug_env = os.environ.get('SPYDER_DEBUG', '')
    if not debug_env.isdigit():
        debug_env = bool(debug_env)
    return int(debug_env)


def debug_print(*message):
    """Output debug messages to stdout"""
    warnings.warn("debug_print is deprecated; use the logging module instead.")
    if get_debug_level():
        ss = STDOUT
        # This is needed after restarting and using debug_print
        for m in message:
            ss.buffer.write(str(m).encode('utf-8'))
        print('', file=ss)


#==============================================================================
# Configuration paths
#==============================================================================
def get_conf_subfolder():
    """Return the configuration subfolder for different ooperating systems."""
    # Spyder settings dir
    # NOTE: During the 2.x.x series this dir was named .spyder2, but
    # since 3.0+ we've reverted back to use .spyder to simplify major
    # updates in version (required when we change APIs by Linux
    # packagers)
    if sys.platform.startswith('openbsd'):
        SUBFOLDER = 'spyder'
    else:
        SUBFOLDER = '.spyder'

    # We can't have PY2 and PY3 settings in the same dir because:
    # 1. This leads to ugly crashes and freezes (e.g. by trying to
    #    embed a PY2 interpreter in PY3)
    # 2. We need to save the list of installed modules (for code
    #    completion) separately for each version
    SUBFOLDER = SUBFOLDER + '-py3'

    # If running a development/beta version, save config in a separate
    # directory to avoid wiping or contaiminating the user's saved stable
    # configuration.
    if use_dev_config_dir():
        SUBFOLDER = SUBFOLDER + '-dev'

    return SUBFOLDER


def get_project_config_folder():
    """Return the default project configuration folder."""
    return '.spyproject'


def get_home_dir():
    """Return user home directory."""
    try:
        # expanduser() returns a raw byte string which needs to be
        # decoded with the codec that the OS is using to represent
        # file paths.
        path = encoding.to_unicode_from_fs(osp.expanduser('~'))
    except Exception:
        path = ''

    if osp.isdir(path):
        return path
    else:
        # Get home from alternative locations
        for env_var in ('HOME', 'USERPROFILE', 'TMP'):
            # os.environ.get() returns a raw byte string which needs to be
            # decoded with the codec that the OS is using to represent
            # environment variables.
            path = encoding.to_unicode_from_fs(os.environ.get(env_var, ''))
            if osp.isdir(path):
                return path
            else:
                path = ''

        if not path:
            raise RuntimeError('Please set the environment variable HOME to '
                               'your user/home directory path so Spyder can '
                               'start properly.')


def get_clean_conf_dir():
    """
    Return the path to a temp clean configuration dir, for tests and safe mode.
    """
    conf_dir = osp.join(
        tempfile.gettempdir(),
        'spyder-clean-conf-dirs',
        CLEAN_DIR_ID,
    )
    return conf_dir


def get_custom_conf_dir():
    """
    Use a custom configuration directory, passed through our command
    line options or by setting the env var below.
    """
    custom_dir = os.environ.get('SPYDER_CONFDIR')
    if custom_dir:
        custom_dir = osp.abspath(custom_dir)

        # Set env var to not lose its value in future calls when the cwd
        # is changed by Spyder.
        os.environ['SPYDER_CONFDIR'] = custom_dir
        return custom_dir


def get_conf_path(filename=None):
    """Return absolute path to the config file with the specified filename."""
    # Define conf_dir
    if running_under_pytest() or get_safe_mode():
        # Use clean config dir if running tests or the user requests it.
        conf_dir = get_clean_conf_dir()
    elif get_custom_conf_dir():
        # Use a custom directory if the user decided to do it through
        # our command line options.
        conf_dir = get_custom_conf_dir()
    elif sys.platform.startswith('openbsd'):
        # This makes us follow the XDG standard to save our settings
        # on Linux, as it was requested on spyder-ide/spyder#2629.
        xdg_config_home = os.environ.get('XDG_CONFIG_HOME', '')
        if not xdg_config_home:
            xdg_config_home = osp.join(get_home_dir(), '.config')

        if not osp.isdir(xdg_config_home):
            os.makedirs(xdg_config_home)

        conf_dir = osp.join(xdg_config_home, get_conf_subfolder())
    else:
        conf_dir = osp.join(get_home_dir(), get_conf_subfolder())

    # Create conf_dir
    if not osp.isdir(conf_dir):
        if running_under_pytest() or get_safe_mode() or get_custom_conf_dir():
            os.makedirs(conf_dir)
        else:
            os.mkdir(conf_dir)

    if filename is None:
        return conf_dir
    else:
        return osp.join(conf_dir, filename)


def get_conf_paths():
    """Return the files that can update system configuration defaults."""
    CONDA_PREFIX = os.environ.get('CONDA_PREFIX', None)

    if os.name == 'nt':
        SEARCH_PATH = (
            'C:/ProgramData/spyder',
        )
    else:
        SEARCH_PATH = (
            '/etc/spyder',
            '/usr/local/etc/spyder',
        )

    if CONDA_PREFIX is not None:
        CONDA_PREFIX = CONDA_PREFIX.replace('\\', '/')
        SEARCH_PATH += (
            '{}/etc/spyder'.format(CONDA_PREFIX),
        )

    SEARCH_PATH += (
        '{}/etc/spyder'.format(sys.prefix),
    )

    if running_under_pytest():
        search_paths = []
        tmpfolder = str(tempfile.gettempdir())
        for i in range(3):
            path = os.path.join(tmpfolder, 'site-config-' + str(i))
            if not os.path.isdir(path):
                os.makedirs(path)
            search_paths.append(path)
        SEARCH_PATH = tuple(search_paths)

    return SEARCH_PATH


def get_module_path(modname):
    """Return module *modname* base path"""
    return osp.abspath(osp.dirname(sys.modules[modname].__file__))


def get_module_data_path(modname, relpath=None, attr_name='DATAPATH'):
    """Return module *modname* data path
    Note: relpath is ignored if module has an attribute named *attr_name*

    Handles py2exe/cx_Freeze distributions"""
    datapath = getattr(sys.modules[modname], attr_name, '')
    if datapath:
        return datapath
    else:
        datapath = get_module_path(modname)
        parentdir = osp.join(datapath, osp.pardir)
        if osp.isfile(parentdir):
            # Parent directory is not a directory but the 'library.zip' file:
            # this is either a py2exe or a cx_Freeze distribution
            datapath = osp.abspath(osp.join(osp.join(parentdir, osp.pardir),
                                            modname))
        if relpath is not None:
            datapath = osp.abspath(osp.join(datapath, relpath))
        return datapath


def get_module_source_path(modname, basename=None):
    """Return module *modname* source path
    If *basename* is specified, return *modname.basename* path where
    *modname* is a package containing the module *basename*

    *basename* is a filename (not a module name), so it must include the
    file extension: .py or .pyw

    Handles py2exe/cx_Freeze distributions"""
    srcpath = get_module_path(modname)
    parentdir = osp.join(srcpath, osp.pardir)
    if osp.isfile(parentdir):
        # Parent directory is not a directory but the 'library.zip' file:
        # this is either a py2exe or a cx_Freeze distribution
        srcpath = osp.abspath(osp.join(osp.join(parentdir, osp.pardir),
                                       modname))
    if basename is not None:
        srcpath = osp.abspath(osp.join(srcpath, basename))
    return srcpath


def is_py2exe_or_cx_Freeze():
    """Return True if this is a py2exe/cx_Freeze distribution of Spyder"""
    return osp.isfile(osp.join(get_module_path('spyder'), osp.pardir))


#==============================================================================
# Translations
#==============================================================================
LANG_FILE = get_conf_path('langconfig')
DEFAULT_LANGUAGE = 'en'

# This needs to be updated every time a new language is added to spyder, and is
# used by the Preferences configuration to populate the Language QComboBox
LANGUAGE_CODES = {
    'en': 'English',
    'fr': 'Français',
    'es': 'Español',
    'hu': 'Magyar',
    'pt_BR': 'Português',
    'ru': 'Русский',
    'zh_CN': '简体中文',
    'ja': '日本語',
    'de': 'Deutsch',
    'pl': 'Polski',
    'fa': 'Persian',
    'hr': 'Croatian',
    'te': 'Telugu',
    'uk': 'Ukrainian',
}

# Disabled languages because their translations are outdated or incomplete
DISABLED_LANGUAGES = ['fa', 'hr', 'hu', 'pl', 'te', 'uk', 'ru']


def get_available_translations():
    """
    List available translations for spyder based on the folders found in the
    locale folder. This function checks if LANGUAGE_CODES contain the same
    information that is found in the 'locale' folder to ensure that when a new
    language is added, LANGUAGE_CODES is updated.
    """
    locale_path = get_module_data_path("spyder", relpath="locale",
                                       attr_name='LOCALEPATH')
    listdir = os.listdir(locale_path)
    langs = [d for d in listdir if osp.isdir(osp.join(locale_path, d))]
    langs = [DEFAULT_LANGUAGE] + langs

    # Remove disabled languages
    langs = list(set(langs) - set(DISABLED_LANGUAGES))

    # Check that there is a language code available in case a new translation
    # is added, to ensure LANGUAGE_CODES is updated.
    retlangs = []
    for lang in langs:
        if lang not in LANGUAGE_CODES:
            if DEV:
                error = ('Update LANGUAGE_CODES (inside config/base.py) if a '
                         'new translation has been added to Spyder. '
                         'Currently missing ' + lang)
                print(error)  # spyder: test-skip
                return ['en']
        else:
            retlangs.append(lang)

    return retlangs


def get_interface_language():
    """
    If Spyder has a translation available for the locale language, it will
    return the version provided by Spyder adjusted for language subdifferences,
    otherwise it will return DEFAULT_LANGUAGE.

    Example:
    1.) Spyder provides ('en', 'de', 'fr', 'es' 'hu' and 'pt_BR'), if the
    locale is either 'en_US' or 'en' or 'en_UK', this function will return 'en'

    2.) Spyder provides ('en', 'de', 'fr', 'es' 'hu' and 'pt_BR'), if the
    locale is either 'pt' or 'pt_BR', this function will return 'pt_BR'
    """

    if os.name == "nt":
        # Changing to locale.getlocale from locale.getdefaultlocale caused some
        # Windows machines to return non BCP47 locale codes. Instead use
        # win32 GetUserDefaultLocaleName which does seem to give BCP47 locale.
        # Fixes spyder-ide/spyder#23318.
        from ctypes import create_unicode_buffer, windll
        bufsize = 85  # LOCALE_NAME_MAX_LENGTH
        buf = create_unicode_buffer(bufsize)
        if windll.kernel32.GetUserDefaultLocaleName(buf, bufsize):
            locale_language = buf.value
        else:
            locale_language = DEFAULT_LANGUAGE
    else:
        # Fixes spyder-ide/spyder#3627.
        try:
            locale_language = locale.getlocale()[0]
        except ValueError:
            locale_language = DEFAULT_LANGUAGE

    # Tests expect English as the interface language
    if running_under_pytest():
        locale_language = DEFAULT_LANGUAGE

    language = DEFAULT_LANGUAGE

    if locale_language is not None:
        spyder_languages = get_available_translations()
        for lang in spyder_languages:
            if locale_language == lang:
                language = locale_language
                break
            elif (locale_language.startswith(lang) or
                    lang.startswith(locale_language)):
                language = lang
                break

    return language


def save_lang_conf(value):
    """Save language setting to language config file"""
    # Needed to avoid an error when trying to save LANG_FILE
    # but the operation fails for some reason.
    # See spyder-ide/spyder#8807.
    try:
        with open(LANG_FILE, 'w') as f:
            f.write(value)
    except EnvironmentError:
        pass


def load_lang_conf():
    """
    Load language setting from language config file if it exists, otherwise
    try to use the local settings if Spyder provides a translation, or
    return the default if no translation provided.
    """
    if osp.isfile(LANG_FILE):
        with open(LANG_FILE, 'r') as f:
            lang = f.read()
    else:
        lang = get_interface_language()
        save_lang_conf(lang)

    # Save language again if it's been disabled
    if lang.strip('\n') in DISABLED_LANGUAGES:
        lang = DEFAULT_LANGUAGE
        save_lang_conf(lang)

    return lang


def get_translation(
    modname: str, dirname: str | None = None
) -> Callable[[str], str]:
    """
    Return the translation callback for module ``modname``.

    Parameters
    ----------
    modname : str
        The module to get the translation callback for.
    dirname : str | None, optional
        The directory name of the module, same as the module name by default.

    Returns
    -------
    Callable[[str], str]
        The gettext translation callback for the given module.
    """
    if dirname is None:
        dirname = modname

    def translate_dumb(x: str) -> str:
        """Dumb function to not use translations."""
        return x

    if building_autodoc():

        def translate_gettext(x: str) -> str:
            """
            Translate a text string to the current language for a module.

            Parameters
            ----------
            x : str
                The string to translate.

            Returns
            -------
            str
                The translated string.
            """
            return x

        return translate_gettext

    locale_path = get_module_data_path(
        dirname, relpath="locale", attr_name="LOCALEPATH"
    )

    # If LANG is defined in Ubuntu, a warning message is displayed,
    # so in Unix systems we define the LANGUAGE variable.
    language = load_lang_conf()
    if os.name == "nt":
        # Trying to set LANG on Windows can fail when Spyder is
        # run with admin privileges.
        # Fixes spyder-ide/spyder#6886.
        try:
            os.environ["LANG"] = language  # Works on Windows
        except Exception:
            return translate_dumb
    else:
        os.environ["LANGUAGE"] = language  # Works on Linux

    if language == "en":
        return translate_dumb

    import gettext

    try:
        _trans = gettext.translation(modname, locale_path)

        def translate_gettext(x: str) -> str:
            """
            Translate a text string to the current language for a module.

            Parameters
            ----------
            x : str
                The string to translate.

            Returns
            -------
            str
                The translated string.
            """
            return _trans.gettext(x)

        return translate_gettext
    except Exception as exc:
        # logging module is not yet initialised at this point
        print(
            f"Could not load translations for {language} due to: "
            f"{exc.__class__.__name__} - {exc}",
            file=sys.stderr,
        )
        return translate_dumb


# Translation callback
_ = get_translation("spyder")


#==============================================================================
# Namespace Browser (Variable Explorer) configuration management
#==============================================================================
# Variable explorer display / check all elements data types for sequences:
# (when saving the variable explorer contents, check_all is True,
CHECK_ALL = False  # XXX: If True, this should take too much to compute...

EXCLUDED_NAMES = ['nan', 'inf', 'infty', 'little_endian', 'colorbar_doc',
                  'typecodes', '__builtins__', '__main__', '__doc__', 'NaN',
                  'Inf', 'Infinity', 'sctypes', 'rcParams', 'rcParamsDefault',
                  'sctypeNA', 'typeNA', 'False_', 'True_']


#==============================================================================
# Conda-based installer application utilities
#==============================================================================
def is_conda_based_app(pyexec=sys.executable):
    """
    Check if Spyder is running from the conda-based installer by looking for
    the `conda_based_app` file.

    If a Python executable is provided, checks if it is in a conda-based
    installer environment or the root environment thereof.
    """
    real_pyexec = osp.realpath(pyexec)  # pyexec may be symlink

    # It seems in some cases the check below detects that a regular conda env
    # corresponds to our installer. This makes things a bit more strict by
    # also validating the env name where Spyder is installed.
    # Fixes spyder-ide/spyder#25549.
    if not "spyder-runtime" in real_pyexec:
        return False

    # Check the presence of conda-based-app file, which is added when creating
    # the installers.
    if os.name == 'nt':
        env_path = osp.dirname(real_pyexec)
    else:
        env_path = osp.dirname(osp.dirname(real_pyexec))

    menu_rel_path = '/Menu/conda-based-app'
    if (
        osp.exists(env_path + menu_rel_path)
        or glob(env_path + '/envs/*' + menu_rel_path)
    ):
        return True
    else:
        return False


def is_installed_all_users():
    """
    Check if conda-based installer is installed for all users.
    Only for conda-based installers.
    """
    real_pyexec = osp.realpath(sys.executable)  # may be symlink

    if not is_conda_based_app(real_pyexec):
        return False

    root = real_pyexec.split("envs")[0]
    return not osp.exists(root + ".nonadmin")


#==============================================================================
# Reset config files
#==============================================================================
SAVED_CONFIG_FILES = ('help', 'onlinehelp', 'path', 'pylint.results',
                      'spyder.ini', 'temp.py', 'temp.spydata', 'template.py',
                      'history.py', 'history_internal.py', 'workingdir',
                      '.projects', '.spyproject', '.ropeproject',
                      'monitor.log', 'monitor_debug.log', 'rope.log',
                      'langconfig', 'spyder.lock',
                      'config{}spyder.ini'.format(os.sep),
                      'config{}transient.ini'.format(os.sep),
                      'lsp_root_path', 'plugins')


def reset_config_files():
    """Remove all config files"""
    print("*** Reset Spyder settings to defaults ***", file=STDERR)
    for fname in SAVED_CONFIG_FILES:
        cfg_fname = get_conf_path(fname)
        if osp.isfile(cfg_fname) or osp.islink(cfg_fname):
            os.remove(cfg_fname)
        elif osp.isdir(cfg_fname):
            shutil.rmtree(cfg_fname)
        else:
            continue
        print("removing:", cfg_fname, file=STDERR)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Standard library imports
import os
from pathlib import Path
import subprocess as sp
import sys

# Local imports
from spyder.config.base import is_conda_based_app


def get_updater_info() -> (str, str):
    """
    Get spyder-updater info

    Returns
    -------
    updater_exe : str
        Full path to the spyder-updater executable. This may not exist.
    version : str
        Version of spyder-updater. If not installed, defaults to "0.0.0"
    """
    real_pyexec = Path(sys.executable).resolve()
    env_dir = real_pyexec.parent.parent
    if os.name == 'nt':
        updater_exe = (
            env_dir / "spyder-updater" / "Scripts" / "spyder-updater.exe"
        )
    else:
        env_dir = env_dir.parent
        updater_exe = env_dir / "spyder-updater" / "bin" / "spyder-updater"

    version = "0.0.0"  # Not installed
    if is_conda_based_app() and updater_exe.exists():
        cmd = " ".join([str(updater_exe), "--version"])
        proc = sp.run(cmd, shell=True, text=True, capture_output=True)
        if proc.returncode == 0:
            version = proc.stdout

    return str(updater_exe), version



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Language Server Protocol introspection configuration tab.
"""

# Standard library imports
import sys

# Third party imports
from qtpy.QtWidgets import QGroupBox, QVBoxLayout

# Local imports
from spyder.api.preferences import SpyderPreferencesTab
from spyder.api.translations import _
from spyder.utils.palette import SpyderPalette


class IntrospectionConfigTab(SpyderPreferencesTab):
    """Introspection settings tab."""

    TITLE = _('Introspection')
    CTRL = "Cmd" if sys.platform == 'darwin' else "Ctrl"

    def __init__(self, parent):
        super().__init__(parent)
        newcb = self.create_checkbox

        introspection_group = QGroupBox(_("Basic features"))
        goto_definition_box = newcb(
            _("Enable Go to definition"),
            'jedi_definition',
            tip=_("If enabled, left-clicking on an object name while \n"
                  "pressing the {} key will go to that object's definition\n"
                  "(if resolved).").format(self.CTRL))
        follow_imports_box = newcb(_("Follow imports when going to a "
                                     "definition"),
                                   'jedi_definition/follow_imports')
        show_signature_box = newcb(_("Show calltips"), 'jedi_signature_help')
        enable_hover_hints_box = newcb(
            _("Enable hover hints"),
            'enable_hover_hints',
            tip=_("If enabled, hovering the mouse pointer over an object\n"
                  "name will display that object's signature and/or\n"
                  "docstring (if present)."))
        introspection_layout = QVBoxLayout()
        introspection_layout.addWidget(goto_definition_box)
        introspection_layout.addWidget(follow_imports_box)
        introspection_layout.addWidget(show_signature_box)
        introspection_layout.addWidget(enable_hover_hints_box)
        introspection_group.setLayout(introspection_layout)

        goto_definition_box.checkbox.toggled.connect(
            follow_imports_box.setEnabled)

        # Advanced group
        advanced_group = QGroupBox(_("Advanced"))
        modules_textedit = self.create_textedit(
            _("Preload the following modules to make completion faster "
              "and more accurate:"),
            'preload_modules'
        )
        modules_textedit.textbox.setStyleSheet(
            f"border: 1px solid {SpyderPalette.COLOR_BACKGROUND_2};"
        )

        advanced_layout = QVBoxLayout()
        advanced_layout.addWidget(modules_textedit)
        advanced_group.setLayout(advanced_layout)

        layout = QVBoxLayout()
        layout.addWidget(introspection_group)
        layout.addWidget(advanced_group)
        layout.addStretch(1)
        self.setLayout(layout)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Third party imports
from qtpy.QtCore import Property, QSize, Signal, Slot
from qtpy.QtGui import QColor, QIcon, QPixmap
from qtpy.QtWidgets import QColorDialog, QHBoxLayout, QLineEdit, QToolButton


class ColorButton(QToolButton):
    """
    Color choosing push button
    """
    colorChanged = Signal(QColor)

    def __init__(self, parent=None):
        QToolButton.__init__(self, parent)
        self.setFixedSize(20, 20)
        self.setIconSize(QSize(12, 12))
        self.clicked.connect(self.choose_color)
        self._color = QColor()

    def choose_color(self):
        color = QColorDialog.getColor(self._color, self.parentWidget(),
                                      'Select Color',
                                      QColorDialog.ShowAlphaChannel)
        if color.isValid():
            self.set_color(color)

    def get_color(self):
        return self._color

    @Slot(QColor)
    def set_color(self, color):
        if color != self._color:
            self._color = color
            self.colorChanged.emit(self._color)
            pixmap = QPixmap(self.iconSize())
            pixmap.fill(color)
            self.setIcon(QIcon(pixmap))

    color = Property("QColor", get_color, set_color)


def text_to_qcolor(text):
    """
    Create a QColor from specified string
    Avoid warning from Qt when an invalid QColor is instantiated
    """
    color = QColor()
    text = str(text)
    if not isinstance(text, str):
        return color
    if text.startswith('#') and len(text)==7:
        correct = '#0123456789abcdef'
        for char in text:
            if char.lower() not in correct:
                return color
    elif text not in list(QColor.colorNames()):
        return color
    color.setNamedColor(text)
    return color


class ColorLayout(QHBoxLayout):
    """Color-specialized QLineEdit layout"""
    def __init__(self, color, parent=None):
        QHBoxLayout.__init__(self)
        assert isinstance(color, QColor)
        self.lineedit = QLineEdit(color.name(), parent)
        fm = self.lineedit.fontMetrics()
        self.lineedit.setMinimumWidth(int(fm.width(color.name()) * 1.2))
        self.lineedit.textChanged.connect(self.update_color)
        self.addWidget(self.lineedit)
        self.colorbtn = ColorButton(parent)
        self.colorbtn.color = color
        self.colorbtn.colorChanged.connect(self.update_text)
        self.addWidget(self.colorbtn)

    def update_color(self, text):
        color = text_to_qcolor(text)
        if color.isValid():
            self.colorbtn.color = color

    def update_text(self, color):
        self.lineedit.setText(color.name())

    def text(self):
        return self.lineedit.text()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------
"""Project creation dialog."""

# Standard library imports
from __future__ import annotations
import os
import os.path as osp
import re
from pathlib import Path
import sys
from typing import TypedDict

# Third party imports
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import (
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
)

# Local imports
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.api.translations import _
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.plugins.projects.api import EmptyProject
from spyder.utils.icon_manager import ima
from spyder.utils.stylesheet import AppStyle, MAC, WIN
from spyder.widgets.config import SpyderConfigPage
from spyder.widgets.sidebardialog import SidebarDialog
from spyder.widgets.helperwidgets import MessageLabel


# =============================================================================
# ---- Auxiliary functions and classes
# =============================================================================
def is_writable(path):
    """
    Check if path has write access.

    Solution taken from https://stackoverflow.com/a/11170037
    """
    filepath = osp.join(path, "__spyder_write_test__.txt")

    try:
        filehandle = open(filepath, 'w')
        filehandle.close()
        os.remove(filepath)
    except (FileNotFoundError, PermissionError):
        return False

    return True


class ValidationReasons(TypedDict):
    missing_info: bool | None
    no_location: bool | None
    location_exists: bool | None
    location_not_writable: bool | None
    spyder_project_exists: bool | None
    wrong_name: bool | None


# =============================================================================
# ---- Pages
# =============================================================================
class BaseProjectPage(SpyderConfigPage, SpyderFontsMixin):
    """Base project page."""

    # SidebarPage API
    MIN_HEIGHT = 300
    MAX_WIDTH = 430 if MAC else (400 if WIN else 420)

    # SpyderConfigPage API
    LOAD_FROM_CONFIG = False

    # Own API
    LOCATION_TEXT = _("Location")
    LOCATION_TIP = None

    def __init__(self, parent):
        super().__init__(parent)

        self._location = self.create_browsedir(
            text=self.LOCATION_TEXT,
            option=None,
            alignment=Qt.Vertical,
            tip=self.LOCATION_TIP,
            status_icon=ima.icon("error"),
        )

        self._validation_label = MessageLabel(self)
        self._validation_label.setVisible(False)

        self._description_font = self.get_font(SpyderFontType.Interface)
        self._description_font.setPointSize(
            self._description_font.pointSize() + 1
        )

    # ---- Public API
    # -------------------------------------------------------------------------
    @property
    def project_location(self):
        """Where the project is going to be created."""
        raise NotImplementedError

    def validate_page(self):
        """Actions to take to validate the page contents."""
        raise NotImplementedError

    @property
    def project_type(self):
        """Project type associated to this page."""
        return EmptyProject

    # ---- Private API
    # -------------------------------------------------------------------------
    def _validate_location(
        self,
        location: str,
        reasons: ValidationReasons | None = None,
        name: str | None = None
    ) -> ValidationReasons:

        if reasons is None:
            reasons: ValidationReasons = {}

        if not location:
            self._location.status_action.setVisible(True)
            self._location.status_action.setToolTip(_("This is empty"))
            reasons["missing_info"] = True
        elif not osp.isdir(location):
            self._location.status_action.setVisible(True)
            self._location.status_action.setToolTip(
                _("This directory doesn't exist")
            )
            reasons["no_location"] = True
        elif not is_writable(location):
            self._location.status_action.setVisible(True)
            self._location.status_action.setToolTip(
                _("This directory is not writable")
            )
            reasons["location_not_writable"] = True
        elif os.name == "nt" and any(
            [re.search(r":", part) for part in Path(location).parts[1:]]
        ):
            # Prevent creating a project in directory with colons.
            # Fixes spyder-ide/spyder#16942
            reasons["wrong_name"] = True
        elif name is not None:
            project_path = osp.join(location, name)
            if os.name == "nt" and re.search(r":", name):
                # Prevent creating a project in directory with colons.
                # Fixes spyder-ide/spyder#16942
                reasons["wrong_name"] = True
            if osp.isdir(project_path):
                reasons["location_exists"] = True
        else:
            spyproject_path = osp.join(location, '.spyproject')
            if osp.isdir(spyproject_path):
                self._location.status_action.setVisible(True)
                self._location.status_action.setToolTip(
                    _("You selected a Spyder project")
                )
                reasons["spyder_project_exists"] = True

        return reasons

    def _compose_failed_validation_text(self, reasons: ValidationReasons):
        n_reasons = list(reasons.values()).count(True)
        prefix = "- " if n_reasons > 1 else ""
        suffix = "<br>" if n_reasons > 1 else ""

        text = ""
        if reasons.get("location_exists"):
            text += (
                prefix
                + _(
                    "The directory you selected for this project already "
                    "exists."
                )
                + suffix
            )
        elif reasons.get("spyder_project_exists"):
            text += (
                prefix
                + _("This directory is already a Spyder project.")
                + suffix
            )
        elif reasons.get("location_not_writable"):
            text += (
                prefix
                + _(
                    "You don't have write permissions in the location you "
                    "selected."
                )
                + suffix
            )
        elif reasons.get("no_location"):
            text += (
                prefix
                + _("The location you selected doesn't exist.")
                + suffix
            )

        if reasons.get("wrong_name"):
            text += (
                prefix
                + _("The directory name you selected is not valid.")
            )

        if reasons.get("missing_info"):
            text += (
                prefix
                + _("There are missing fields on this page.")
            )

        return text


class NewDirectoryPage(BaseProjectPage):
    """New directory project page."""

    LOCATION_TIP = _(
        "Select the location where the project directory will be created"
    )
    PROJECTS_DOCS_URL = (
        "https://docs.spyder-ide.org/current/panes/projects.html"
    )

    def get_name(self):
        return _("New directory")

    def get_icon(self):
        return self.create_icon("folder_new")

    def setup_page(self):
        description = QLabel(_("Start a project in a new directory"))
        description.setWordWrap(True)
        description.setFont(self._description_font)

        docs_reference = QLabel(
            _(
                "To learn more about projects, see our "
                '<a href="{0}">documentation</a>.'
            ).format(self.PROJECTS_DOCS_URL)
        )
        docs_reference.setOpenExternalLinks(True)

        self._name = self.create_lineedit(
            text=_("Project directory"),
            option=None,
            tip=_(
                "A directory with this name will be created in the location "
                "below"
            ),
            status_icon=ima.icon("error"),
        )

        layout = QVBoxLayout()
        layout.addWidget(description)
        layout.addWidget(docs_reference)
        layout.addSpacing(5 * AppStyle.MarginSize)
        layout.addWidget(self._name)
        layout.addSpacing(5 * AppStyle.MarginSize)
        layout.addWidget(self._location)
        layout.addSpacing(7 * AppStyle.MarginSize)
        layout.addWidget(self._validation_label)
        layout.addStretch()
        self.setLayout(layout)

    @property
    def project_location(self):
        return osp.normpath(
            osp.join(self._location.textbox.text(), self._name.textbox.text())
        )

    def validate_page(self):
        name = self._name.textbox.text()

        # Avoid using "." as location, which is the result of os.normpath("")
        location_text = self._location.textbox.text()
        location = osp.normpath(location_text) if location_text else ""

        # Clear validation state
        self._validation_label.setVisible(False)
        for widget in [self._name, self._location]:
            widget.status_action.setVisible(False)

        # Perform validation
        reasons: ValidationReasons = {}
        if not name:
            self._name.status_action.setVisible(True)
            self._name.status_action.setToolTip(_("This is empty"))
            reasons["missing_info"] = True

        reasons = self._validate_location(location, reasons, name)
        if reasons:
            if reasons.get("location_exists"):
                self._name.status_action.setVisible(True)
                self._name.status_action.setToolTip(
                    _("A directory with this name already exists")
                )
            if reasons.get("wrong_name"):
                self._name.status_action.setVisible(True)
                self._name.status_action.setToolTip(
                    _("The project directory can't contain ':'")
                )
            self._validation_label.set_text(
                self._compose_failed_validation_text(reasons)
            )
            self._validation_label.setVisible(True)

        return False if reasons else True


class ExistingDirectoryPage(BaseProjectPage):
    """Existing directory project page."""

    LOCATION_TEXT = _("Project path")
    LOCATION_TIP = _("Select the directory to use for the project")

    def get_name(self):
        return _("Existing directory")

    def get_icon(self):
        return self.create_icon("DirClosedIcon")

    def setup_page(self):
        description = QLabel(
            _("Create a Spyder project in an existing directory")
        )
        description.setWordWrap(True)
        description.setFont(self._description_font)

        layout = QVBoxLayout()
        layout.addWidget(description)
        layout.addSpacing(5 * AppStyle.MarginSize)
        layout.addWidget(self._location)
        layout.addSpacing(7 * AppStyle.MarginSize)
        layout.addWidget(self._validation_label)
        layout.addStretch()
        self.setLayout(layout)

    @property
    def project_location(self):
        return osp.normpath(self._location.textbox.text())

    def validate_page(self):
        # Clear validation state
        self._validation_label.setVisible(False)
        self._location.status_action.setVisible(False)

        # Avoid using "." as location, which is the result of os.normpath("")
        location_text = self._location.textbox.text()
        location = osp.normpath(location_text) if location_text else ""

        # Perform validation
        reasons = self._validate_location(location)
        if reasons:
            self._validation_label.set_text(
                self._compose_failed_validation_text(reasons)
            )
            self._validation_label.setVisible(True)

        return False if reasons else True


# =============================================================================
# ---- Dialog
# =============================================================================
class ProjectDialog(SidebarDialog):
    """Project creation dialog."""

    FIXED_SIZE = True
    MIN_WIDTH = 740 if MAC else (670 if WIN else 730)
    MIN_HEIGHT = 470 if MAC else (420 if WIN else 450)
    PAGES_MINIMUM_WIDTH = 450
    PAGE_CLASSES = [NewDirectoryPage, ExistingDirectoryPage]

    sig_project_creation_requested = Signal(str, str, object)
    """
    This signal is emitted to request the Projects plugin the creation of a
    project.

    Parameters
    ----------
    project_path: str
        Location of project.
    project_type: str
        Type of project as defined by project types.
    project_packages: object
        Package to install. Currently not in use.
    """

    def __init__(self, parent):
        """Project creation dialog."""
        super().__init__(parent=parent)
        self.project_data = {}

        self.setWindowFlags(
            self.windowFlags() & ~Qt.WindowContextHelpButtonHint
        )
        self.setWindowTitle(_('Create new project'))
        self.setWindowIcon(ima.icon("project_new"))

    def create_buttons(self):
        bbox = SpyderDialogButtonBox(
            QDialogButtonBox.Cancel, orientation=Qt.Horizontal
        )

        self.button_create = QPushButton(_('Create'))
        self.button_create.clicked.connect(self.create_project)
        bbox.addButton(self.button_create, QDialogButtonBox.ActionRole)

        layout = QHBoxLayout()
        layout.addWidget(bbox)
        return bbox, layout

    def create_project(self):
        """Create project."""
        page = self.get_page()

        # Validate info
        if not page.validate_page():
            return

        self.project_data = {
            "root_path": osp.normpath(page.project_location),
            "project_type": page.project_type.ID,
        }
        self.sig_project_creation_requested.emit(
            page.project_location,
            page.project_type.ID,
            [],
        )
        self.accept()


def test():
    """Local test."""
    from spyder.utils.qthelpers import qapplication
    from spyder.config.base import running_in_ci

    app = qapplication()

    dlg = ProjectDialog(None)

    if not running_in_ci():
        from spyder.utils.stylesheet import APP_STYLESHEET
        app.setStyleSheet(str(APP_STYLESHEET))

    dlg.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-

# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Update Manager widgets."""

# Standard library imports
import json
import logging
import os
import os.path as osp
import subprocess
import sys
from sysconfig import get_path

# Third-party imports
from qtpy.QtCore import Qt, QThread, QTimer, Signal
from qtpy.QtWidgets import (
    QGridLayout,
    QMessageBox,
    QProgressBar,
    QPushButton,
    QTextEdit,
    QWidget,
)
from spyder_kernels.utils.pythonenv import is_conda_env

# Local imports
from spyder import __version__
from spyder.api.config.mixins import SpyderConfigurationAccessor
from spyder.api.fonts import SpyderFontsMixin, SpyderFontType
from spyder.api.translations import _
from spyder.config.base import is_conda_based_app, is_installed_all_users
from spyder.config.gui import is_dark_interface
from spyder.plugins.updatemanager.workers import (
    UpdateType,
    validate_download,
    WorkerUpdate,
    WorkerUpdateUpdater,
    WorkerDownloadInstaller
)
from spyder.plugins.updatemanager.utils import get_updater_info
from spyder.utils.conda import find_conda, is_anaconda_pkg
from spyder.utils.palette import SpyderPalette
from spyder.utils.programs import (
    get_temp_dir,
    is_program_installed,
    find_program
)
from spyder.widgets.helperwidgets import MessageCheckBox

# Logger setup
logger = logging.getLogger(__name__)

# Update manager process statuses
NO_STATUS = __version__
UPDATING_UPDATER = _("Updating Spyder-updater")
DOWNLOADING_INSTALLER = _("Downloading update")
DOWNLOAD_FINISHED = _("Download finished")
PENDING = _("Update available")
CHECKING = _("Checking for updates")
INSTALL_ON_CLOSE = _("Install on close")

HEADER = _("<h3>Spyder {} is available!</h3><br>")
URL_I = 'https://docs.spyder-ide.org/current/installation.html'

SKIP_CHECK_UPDATE = (
    sys.executable.startswith(('/usr/bin/', '/usr/local/bin/'))
    or (
        not is_conda_env(sys.prefix)
        and osp.exists(osp.join(get_path('stdlib'), 'EXTERNALLY-MANAGED'))
    )
    or sys.platform not in ('openbsd', 'darwin', 'win32')  # Supported platforms
)


class UpdateManagerWidget(QWidget, SpyderConfigurationAccessor):
    """Check for updates widget."""

    CONF_SECTION = "update_manager"

    sig_disable_actions = Signal(bool)
    """
    Signal to disable plugin actions during check for update.

    Parameters
    ----------
    disable: bool
        True to disable, False to re-enable.
    """

    sig_block_status_signals = Signal(bool)
    """
    Signal to block signals from update manager status during
    check for update.

    Parameters
    ----------
    block: bool
        True to block, False to unblock.
    """

    sig_download_progress = Signal(int)
    """
    Signal to send the download progress.

    Parameters
    ----------
    percent_progress: int
        Percent of the data downloaded until now.
    """

    sig_set_status = Signal(str, str)
    """
    Signal to set the status of update manager.

    Parameters
    ----------
    status: str
        Status string.
    latest_release: str
        Latest release version detected.
    """

    sig_exception_occurred = Signal(dict)
    """
    Pass untracked exceptions from workers to error reporter.
    """

    sig_install_on_close = Signal(bool)
    """
    Signal to request running the install process on close.

    Parameters
    ----------
    install_on_close: bool
        Whether to install on close.
    """

    sig_quit_requested = Signal()
    """
    This signal can be emitted to request the main application to quit.
    """

    def __init__(self, parent):
        super().__init__(parent)

        self.startup = None
        self.update_thread = None
        self.update_worker = None
        self.update_timer = None
        self.asset_info = None

        self.update_updater_thread = None
        self.update_updater_worker = None

        self.cancelled = False
        self.download_thread = None
        self.download_worker = None
        self.progress_dialog = None
        self.installer_path = None

        self.restart_spyder = True

    # ---- General

    def set_status(self, status=NO_STATUS):
        """Set the update manager status."""
        version = None
        if self.asset_info is not None:
            version = self.asset_info["version"]
        self.sig_set_status.emit(status, str(version))

    def cleanup_threads(self):
        """Clean up QThreads"""
        if self.update_timer is not None:
            self.update_timer.stop()
        if self.update_thread is not None:
            self.update_thread.quit()
            self.update_thread.wait()
        if self.download_thread is not None:
            self.download_worker.cancelled = True
            self.download_thread.quit()
            self.download_thread.wait()

    def handle_exception(self, exc):
        """Cleanup if exception occurs"""
        if self.progress_dialog is not None:
            self.progress_dialog.accept()
            self.progress_dialog = None

        self.cleanup_threads()

        self.set_status(NO_STATUS)

        self.sig_exception_occurred.emit(exc)

    # ---- Check Update

    def start_check_update(self, startup=False):
        """
        Check for Spyder updates using a QThread.

        Do not check for updates if the environment is a system or a managed
        environment.

        Update actions are disabled in the menubar and statusbar while
        checking for updates.

        If startup is True, then checking for updates is delayed 1 min;
        actions are disabled during this time as well.
        """
        if SKIP_CHECK_UPDATE:
            logger.debug(
                "Skip check for updates: system or managed environment."
            )
            return

        logger.debug(f"Checking for updates. startup = {startup}.")

        # Disable check_update_action while the thread is working
        self.sig_disable_actions.emit(True)

        self.startup = startup
        self.cleanup_threads()

        self.update_thread = QThread(None)
        self.update_worker = WorkerUpdate(self.get_conf('check_stable_only'))
        self.update_worker.sig_exception_occurred.connect(
            self.handle_exception
        )
        self.update_worker.sig_ready.connect(self._process_check_update)
        self.update_worker.sig_ready.connect(self.update_thread.quit)
        self.update_worker.sig_ready.connect(
            lambda: self.sig_disable_actions.emit(False)
        )
        self.update_worker.moveToThread(self.update_thread)
        self.update_thread.started.connect(lambda: self.set_status(CHECKING))
        self.update_thread.started.connect(self.update_worker.start)

        # Delay starting this check to avoid blocking the main window
        # while loading.
        # Fixes spyder-ide/spyder#15839
        if self.startup:
            self.update_timer = QTimer(self)
            self.update_timer.setInterval(60000)
            self.update_timer.setSingleShot(True)
            self.sig_block_status_signals.emit(True)
            self.update_timer.timeout.connect(
                lambda: self.sig_block_status_signals.emit(False)
            )
            self.update_timer.timeout.connect(self.update_thread.start)
            self.update_timer.start()
        else:
            # Otherwise, start immediately
            self.update_thread.start()

    def _process_check_update(self):
        """Process the results of check update."""
        # Get results from worker
        update_available = self.update_worker.asset_info is not None
        error_msg = self.update_worker.error
        checkbox = self.update_worker.checkbox

        # Always set status, regardless of error, DEV, or startup
        self.set_status(PENDING if update_available else NO_STATUS)

        # self.startup = True is used on startup, so only positive feedback is
        # given. self.startup = False is used after startup when using the menu
        # action, and gives feeback if updates are or are not found.
        if (
            self.startup and           # startup and...
            (error_msg is not None     # there is an error
             or not update_available)  # or no updates available
        ):
            # Do not alert the user to anything
            pass
        elif error_msg is not None:
            error_messagebox(self, error_msg, checkbox)
        elif update_available:
            self.start_update()
        else:
            info_messagebox(self, _("Spyder is up to date."), checkbox=True)

    # ---- Download Update

    def _set_installer_path(self):
        dirname = osp.join(
            get_temp_dir(), 'updates', str(self.asset_info["version"])
        )
        self.installer_path = osp.join(dirname, self.asset_info['filename'])

        logger.info(f"Update type: {self.asset_info['update_type']}")

    def _validate_download(self):
        update_downloaded = False
        if osp.exists(self.installer_path):
            update_downloaded = validate_download(
                self.installer_path, self.asset_info["checksum"]
            )

        logger.debug(f"Update already downloaded: {update_downloaded}")

        return update_downloaded

    def start_update(self):
        """
        Start the update process

        Request input from user whether to download the installer; upon
        affirmation, proceed with download then to confirm install.

        If the installer is already downloaded, proceed to confirm install.
        """
        self.asset_info = self.update_worker.asset_info
        self._set_installer_path()
        version = self.asset_info["version"]

        if is_conda_based_app():
            # Only perform updates for the standalone installers. Otherwise,
            # users find the process confusing and it leads to weird errors
            # when updating in conda envs.
            # See spyder-ide/spyder#25560, spyder-ide/spyder#25549 and
            # spyder-ide/spyder#25533
            if self._validate_download():
                if self.asset_info["update_type"] == UpdateType.Major:
                    # Major updates don't need Updater, start install
                    self.set_status(DOWNLOAD_FINISHED)
                    self._confirm_install()
                else:
                    # Minor/micro updates need the Updater
                    self._start_update_updater()
            else:
                msg = _("Would you like to download the update?")
                box = confirm_messagebox(
                    self,
                    msg,
                    _("Spyder update"),
                    version=version,
                    checkbox=True,
                )
                if box.result() == QMessageBox.Yes:
                    if self.asset_info["update_type"] == UpdateType.Major:
                        self._start_download()
                    else:
                        self._start_update_updater()
        else:
            # For any other type of installation, simply tell users the command
            # they need to run to perform the update.
            manual_update_messagebox(self, version, self.update_worker.channel)

    def _start_update_updater(self):
        """Check for and install updates for Spyder-updater."""
        self.sig_disable_actions.emit(True)
        self.set_status(UPDATING_UPDATER)

        self.progress_dialog = ProgressDialog(
            self, _("Updating Spyder's updater..."),
            cancel_btn=False
        )
        # Show progress bar as busy
        self.progress_dialog.update_progress(0, 0)

        self.update_updater_thread = QThread(None)
        self.update_updater_worker = WorkerUpdateUpdater(
            self.get_conf('check_stable_only')
        )
        self.update_updater_worker.sig_exception_occurred.connect(
            self.handle_exception
        )

        self.update_updater_worker.sig_ready.connect(
            self._process_update_updater
        )
        self.update_updater_worker.sig_ready.connect(
            self.update_updater_thread.quit
        )
        self.update_updater_worker.sig_ready.connect(
            lambda: self.sig_disable_actions.emit(False)
        )
        self.update_updater_worker.moveToThread(self.update_updater_thread)
        self.update_updater_thread.started.connect(
            self.update_updater_worker.start
        )
        self.update_updater_thread.start()

    def _process_update_updater(self):
        """Process possible errors when updating the updater"""
        error = self.update_updater_worker.error
        if error is None:
            self._start_download()
            return

        self.set_status(PENDING)
        if self.progress_dialog is not None:
            self.progress_dialog.accept()
            self.progress_dialog = None

        if isinstance(error, subprocess.CalledProcessError):
            error_msg = _("Error updating Spyder-updater.")
            details = [
                "*** COMMAND ***",
                error.cmd.strip(),
                "\n*** STDOUT ***",
                error.output.strip(),
                "\n*** STDERR ***",
                error.stderr.strip(),
            ]
            detailed_error_messagebox(
                self, error_msg, details="\n".join(details)
            )

    def _start_download(self):
        """
        Start downloading the installer in a QThread
        and set downloading status.
        """
        if self._validate_download():
            # Update already downloaded, start install
            self._confirm_install()
            return

        self.cancelled = False
        if self.progress_dialog is not None:
            self.progress_dialog.accept()
            self.progress_dialog = None

        self.download_worker = WorkerDownloadInstaller(
            self.asset_info, self.installer_path
        )

        self.sig_disable_actions.emit(True)
        self.set_status(DOWNLOADING_INSTALLER)

        # Only show progress bar for installers
        if not self.installer_path.endswith('zip'):
            self.progress_dialog = ProgressDialog(
                self,
                _("Downloading Spyder {} ...").format(
                    self.asset_info["version"]
                )
            )
            self.progress_dialog.cancel.clicked.connect(self._cancel_download)

        self.download_thread = QThread(None)
        self.download_worker.sig_exception_occurred.connect(
            self.handle_exception
        )
        self.download_worker.sig_ready.connect(self._confirm_install)
        self.download_worker.sig_ready.connect(self.download_thread.quit)
        self.download_worker.sig_ready.connect(
            lambda: self.sig_disable_actions.emit(False)
        )
        self.download_worker.sig_download_progress.connect(
            self._update_download_progress
        )
        self.download_worker.moveToThread(self.download_thread)
        self.download_thread.started.connect(self.download_worker.start)
        self.download_thread.start()

    def show_progress_dialog(self):
        """Show download progress if previously hidden"""
        if self.progress_dialog is not None:
            if not self.progress_dialog.isVisible():
                self.progress_dialog.show()
            else:
                self.progress_dialog.hide()

    def _update_download_progress(self, progress, total):
        """Update download progress in dialog and status bar"""
        if self.progress_dialog is not None:
            self.progress_dialog.update_progress(progress, total)

        percent_progress = 0
        if total > 0:
            percent_progress = round((progress / total) * 100)
        self.sig_download_progress.emit(percent_progress)

    def _cancel_download(self):
        """Cancel the installation in progress."""
        self.download_worker.paused = True
        msg = _('Do you really want to cancel the download?')
        box = confirm_messagebox(
            self, msg, _('Spyder download'), critical=True
        )
        if box.result() == QMessageBox.Yes:
            self.cancelled = True
            self.cleanup_threads()
            self.set_status(PENDING)
        else:
            self.progress_dialog.show()
            self.download_worker.paused = False

    def _confirm_install(self):
        """
        Ask users if they want to proceed with the install immediately
        or on close.
        """
        if self.progress_dialog is not None:
            self.progress_dialog.accept()

        if self.cancelled:
            return

        if self.download_worker and self.download_worker.error:
            # If download error, do not proceed with install
            if self.progress_dialog is not None:
                self.progress_dialog.reject()
            self.set_status(PENDING)
            error_messagebox(self, self.download_worker.error)
            return

        if self.download_worker:
            self.set_status(DOWNLOAD_FINISHED)

        msg = _("Would you like to install it?")
        box = confirm_messagebox(
            self,
            msg,
            _('Spyder install'),
            version=self.asset_info["version"],
            on_close=True
        )

        if box.result() == QMessageBox.Yes:
            self.restart_spyder = True
            self.sig_install_on_close.emit(True)
            self.set_conf("update_performed", True)
            self.sig_quit_requested.emit()
        elif box.result() == 0:  # 0 is result of 3rd push-button
            self.restart_spyder = False
            self.set_conf("update_performed", True)
            self.sig_install_on_close.emit(True)
            self.set_status(INSTALL_ON_CLOSE)

    def start_install(self):
        """Install from downloaded installer or update through conda."""
        if self.asset_info["update_type"] == UpdateType.Major:
            self._start_major_install()
        else:
            self._start_updater()

    def _start_major_install(self):
        """
        Install major update from downloaded installer.

        Major updates are performed using the installers directly.
        macOS and Windows installers have GUI interfaces; Linux requires
        a terminal for the GUI.
        """
        if os.name == 'nt':
            cmd = ['start', '"Update Spyder"']
        elif sys.platform == 'darwin':
            cmd = ["open"]
        else:
            programs = [
                {'cmd': 'gnome-terminal', 'exe-opt': '--window --'},
                {'cmd': 'konsole', 'exe-opt': '-e'},
                {'cmd': 'xfce4-terminal', 'exe-opt': '-x'},
                {'cmd': 'xterm', 'exe-opt': '-e'}
            ]
            for program in programs:
                if is_program_installed(program['cmd']):
                    cmd = [program['cmd'], program['exe-opt']]
                    break

        cmd.append(self.installer_path)

        env = os.environ.copy()
        if sys.platform == "darwin":
            # PKG installers will not pass environment variables in the GUI.
            # To communicate whether to start Spyder, create dummy file.
            no_start_file = osp.join(
                osp.dirname(self.installer_path), "no-start-spyder"
            )
            if self.restart_spyder and osp.exists(no_start_file):
                os.remove(no_start_file)
            if not self.restart_spyder:
                open(no_start_file, 'w').close()
        else:
            # EXE and SH installers will pass environment variables
            env.update({"START_SPYDER": str(self.restart_spyder)})

        logger.debug(f"Restart Spyder after install: {self.restart_spyder}")

        logger.debug(f"""Update command: "{' '.join(cmd)}" """)

        subprocess.Popen(' '.join(cmd), shell=True, env=env)

    def _start_updater(self):
        """
        Start updater application.

        For minor/micro updates, Spyder Updater provides the GUI showing
        progress for updating the runtime environment.
        """
        if self.get_conf('high_dpi_custom_scale_factor', section='main'):
            scale_factors = self.get_conf(
                'high_dpi_custom_scale_factors',
                section='main'
            )
            scale_factor = float(scale_factors.split(";")[0])
        else:
            scale_factor = 1

        info = {
            "install_file": self.installer_path,
            "conda_exec": find_conda(),
            "env_path": sys.prefix,
            "update_type": self.asset_info["update_type"],
            "window_title": _("Spyder update"),
            "scale_factor": scale_factor,
            "initial_message": _(
                "Updating Spyder, this will take a few minutes ..."
            ),
            "success_message": _(
                "The update was succesful! Spyder will be launched shortly"
            ),
            "failure_message": _("Unfortunately the update failed"),
            "error_message": _("There was an error in the update process"),
            "details_title": _("Show details"),
            "font_family": self.get_conf(
                "app_font/family", section="appearance"
            ),
            "font_size": int(
                self.get_conf("app_font/size", section="appearance")
            ),
            "monospace_font_family": self.get_conf(
                "monospace_app_font/family", section="appearance"
            ),
            "monospace_font_size": int(
                self.get_conf("monospace_app_font/size", section="appearance")
            ),
            "interface_theme": "dark" if is_dark_interface() else "light",
            "icon_color": SpyderPalette.ICON_1,
        }

        info_file = osp.join(
            osp.dirname(self.installer_path), "update-info.json"
        )
        with open(info_file, "w") as f:
            json.dump(info, f, indent=4)

        # Launch updater
        updater_path, __ = get_updater_info()
        cmd = [updater_path, "--update-info-file", info_file]
        if self.restart_spyder:
            cmd.append("--start-spyder")

        kwargs = dict(shell=True)
        if os.name == "nt" and is_installed_all_users():
            # Elevate UAC
            kwargs.update(executable=find_program("powershell"))
            cmd = [
                "start",
                "-FilePath",
                f'"{updater_path}"',
                "-ArgumentList",
                ",".join([f"'{a}'" for a in cmd[1:]]),
                "-WindowStyle",
                "Hidden",
                "-Verb",
                "RunAs",
            ]

        subprocess.Popen(" ".join(cmd), **kwargs)


class UpdateMessageBox(QMessageBox):
    def __init__(self, icon=None, text=None, parent=None):
        super().__init__(icon=icon, text=text, parent=parent)
        self.setWindowModality(Qt.NonModal)
        self.setTextFormat(Qt.RichText)


class DetailedUpdateMessageBox(UpdateMessageBox, SpyderFontsMixin):
    def __init__(self, icon=None, text=None, parent=None, details=None):
        super().__init__(icon=icon, text=text, parent=parent)
        self.setSizeGripEnabled(True)
        self.details = None
        self.setDetailedText(details)

    def setDetailedText(self, details=None):
        """
        Override setDetailedText.

        Note: It is critical that QGridLayout.setRowStretch is called after
        QMessageBox.setDetailedText in order for the stretch behavior to work
        properly. That is the primary reason for overriding setDetailedText.
        """
        if self.details is not None:
            self.details.setText(details)
            return

        super().setDetailedText(details)
        self.details = self.findChild(QTextEdit)

        self.details.setFont(self.get_font(SpyderFontType.Monospace))
        self.details.setLineWrapMode(self.details.NoWrap)
        self.details.setMinimumSize(400, 110)
        self.details.setLineWrapMode(0)

        qgl = self.findChild(QGridLayout)
        qgl.setRowStretch(1, 0)
        qgl.setRowStretch(3, 100)  # QTextEdit should take all the stretch

    def event(self, event):
        """Override to allow resizing the dialog when details are visible."""
        if event.type() in (event.LayoutRequest, event.Resize):
            if event.type() == event.Resize:
                result = super().event(event)
            else:
                result = False

            # Allow resize only if details is available and visible.
            if self.details and self.details.isVisible():
                self.details.setMaximumSize(10000, 10000)
                self.setMaximumSize(10000, 10000)

            return result
        return super().event(event)


class UpdateMessageCheckBox(MessageCheckBox):
    def __init__(self, icon=None, text=None, parent=None):
        super().__init__(icon=icon, text=text, parent=parent)
        self.setTextFormat(Qt.RichText)
        self._parent = parent
        self.set_checkbox_text(_("Check for updates at startup"))
        self.option = 'check_updates_on_startup'
        self.accepted.connect(self.on_close)
        self.rejected.connect(self.on_close)
        if self._parent is not None:
            self.set_checked(parent.get_conf(self.option))

    def on_close(self):
        if self._parent is not None:
            self._parent.set_conf(self.option, self.is_checked())


class ProgressDialog(UpdateMessageBox):
    """Update progress installation dialog."""

    def __init__(self, parent, text, cancel_btn=True):
        super().__init__(icon=QMessageBox.NoIcon, text=text, parent=parent)
        self.setWindowTitle(_("Spyder update"))

        self._progress_bar = QProgressBar(self)
        self._progress_bar.setMinimumWidth(250)
        self._progress_bar.setFixedHeight(15)

        layout = self.layout()
        layout.addWidget(self._progress_bar, 1, 1)

        self.okay = QPushButton(_("OK"))
        self.addButton(self.okay, QMessageBox.YesRole)
        if cancel_btn:
            self.cancel = QPushButton(_("Cancel"))
            self.addButton(self.cancel, QMessageBox.NoRole)
        self.setDefaultButton(self.okay)

        self.show()

    def update_progress(self, progress, total):
        """Update installation progress bar."""
        self._progress_bar.setMaximum(total)
        self._progress_bar.setValue(progress)


def error_messagebox(parent, error_msg, checkbox=False):
    # Use a message box with a checkbox to disable updates when required, or a
    # standard one otherwise.
    box_class = UpdateMessageCheckBox if checkbox else UpdateMessageBox
    box = box_class(icon=QMessageBox.Warning, text=error_msg, parent=parent)
    box.setWindowTitle(_("Spyder update error"))
    box.show()
    return box


def detailed_error_messagebox(parent, msg, details):
    box = DetailedUpdateMessageBox(
        icon=QMessageBox.Warning, text=msg, parent=parent, details=details
    )
    box.show()
    return box


def info_messagebox(parent, message, version=None, checkbox=False):
    box_class = UpdateMessageCheckBox if checkbox else UpdateMessageBox
    message = HEADER.format(version) + message if version else message
    box = box_class(icon=QMessageBox.Information, text=message, parent=parent)
    box.setWindowTitle(_("New Spyder version"))
    box.show()
    return box


def confirm_messagebox(parent, message, title, version=None, critical=False,
                       checkbox=False, on_close=False):
    box_class = UpdateMessageCheckBox if checkbox else UpdateMessageBox
    message = HEADER.format(version) + message if version else message
    box = box_class(
        icon=QMessageBox.Critical if critical else QMessageBox.Question,
        text=message, parent=parent
    )
    box.setWindowTitle(title)
    box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
    box.setDefaultButton(QMessageBox.Yes)
    if on_close:
        box.addButton(_("After closing"), QMessageBox.YesRole)
    box.exec()
    return box


def manual_update_messagebox(parent, latest_release, channel):
    msg = ""
    if os.name == "nt":
        if is_conda_env(sys.prefix):
            msg += _("Run the following command or commands in "
                     "the Anaconda prompt to update manually:"
                     "<br><br>")
        else:
            msg += _("Run the following command in a cmd prompt "
                     "to update manually:<br><br>")
    else:
        if is_conda_env(sys.prefix):
            msg += _("Run the following command or commands in a "
                     "terminal to update manually:<br><br>")
        else:
            msg += _("Run the following command in a terminal to "
                     "update manually:<br><br>")

    if is_conda_env(sys.prefix):
        is_pypi = channel == 'pypi'

        if is_anaconda_pkg() and not is_pypi:
            msg += "<code>conda update anaconda</code><br>"

        if is_pypi:
            dont_mix_pip_conda_video = (
                "https://youtu.be/Ul79ihg41Rs"
            )

            msg += (
                "<code>pip install --upgrade spyder</code>"
                "<br><br><br>"
            )

            msg += _(
                "<b>Important note:</b> You installed Spyder with "
                "pip in a Conda environment, which is not a good "
                "idea. See <a href=\"{}\">our video</a> for more "
                "details about it."
            ).format(dont_mix_pip_conda_video)
        else:
            if channel == 'pkgs/main':
                channel = '-c defaults'
            elif channel is not None:
                channel = f'-c {channel}'
            else:
                channel = ''

            msg += (
                f"<code>conda install {channel} "
                f"spyder={latest_release}"
                f"</code><br><br><br>"
            )

            msg += _(
                "<b>Important note:</b> Since you installed "
                "Spyder with Anaconda, please don't use pip "
                "to update it as that will break your "
                "installation."
            )
    else:
        msg += "<code>pip install --upgrade spyder</code><br>"

    msg += _(
        "<br><br>For more information, visit our "
        "<a href=\"{}\">installation guide</a>."
    ).format(URL_I)
    return info_messagebox(parent, msg)



# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Main widget for plugins showing content from the IPython Console.

Used in, for example, the :guilabel:`Variable Explorer`, :guilabel:`Plots`,
:guilabel:`Debugger` and :guilabel:`Profiler` plugins.
"""

from __future__ import annotations

# Standard library imports
from typing import TYPE_CHECKING, Any

# Third party imports
from qtpy.QtWidgets import QStackedWidget, QVBoxLayout

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.main_widget import PluginMainWidget
from spyder.widgets.emptymessage import EmptyMessageWidget

if TYPE_CHECKING:
    from qtpy.QtWidgets import QWidget

    import spyder.plugins.ipythonconsole.widgets


class _ErroredMessageWidget(EmptyMessageWidget):
    """Widget to show when the kernel's shell failed to start."""

    def __init__(
        self,
        parent: ShellConnectMainWidget,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Initialize :class:`EmptyMessageWidget` with content to show for errors.

        Parameters
        ----------
        parent : ShellConnectMainWidget
            The parent widget of this one.
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget that failed to launch.

        Returns
        -------
        None
        """
        super().__init__(
            parent,
            icon_filename=(
                "console-remote-off"
                if shellwidget.is_remote()
                else "console-off"
            ),
            text=_("No connected console"),
            description=_(
                "The current console has no active kernel, so there is no "
                "content to show here"
            ),
            adjust_on_resize=True,
        )

        self.is_empty: bool = False
        """
        If the current widget has no content to show.

        Used to show this widget if :class:`ShellConnectMainWidget` has an
        empty message.
        """


class ShellConnectMainWidget(PluginMainWidget):
    """
    Main widget to use in a plugin that shows different content per-console.

    It is composed of a :class:`QStackedWidget` to stack the widget associated
    to each console shell widget, and only show one of them at a time.
    The current widget in the stack displays the content corresponding to
    the console with focus.

    Used in the :guilabel:`Variable Explorer` and :guilabel:`Plots` plugins,
    for example, to show the variables and plots of the current console.
    """

    def __init__(
        self, *args: Any, set_layout: bool = True, **kwargs: Any
    ) -> None:
        """
        Create a new main widget to show console-specific content.

        Parameters
        ----------
        *args : Any
            Positional arguments to pass to
            :meth:`PluginMainWidget.__init__() <spyder.api.widgets.main_widget.PluginMainWidget.__init__>`.
        set_layout : bool, optional
            Whether to add the created widget to a layout, ``True`` by default.
        **kwargs : Any
            Keyword arguments to pass to
            :meth:`PluginMainWidget.__init__() <spyder.api.widgets.main_widget.PluginMainWidget.__init__>`.

        Returns
        -------
        None
        """
        super().__init__(*args, **kwargs)

        # Widgets
        if not (
            self.SHOW_MESSAGE_WHEN_EMPTY
            and self.get_conf(
                "show_message_when_panes_are_empty", section="main"
            )
        ):
            self._stack = QStackedWidget(self)

            if set_layout:
                layout = QVBoxLayout()
                layout.addWidget(self._stack)
                self.setLayout(layout)

        self._shellwidgets = {}

    # ---- PluginMainWidget API
    # ------------------------------------------------------------------------
    def current_widget(self) -> QWidget:
        """
        Get the currently displayed widget (either the stack or error widget).

        Returns
        -------
        QWidget
            The currently displayed widget, either the active widget in the
            stack associated with the current
            :class:`~spyder.plugins.ipythonconsole.widgets.ShellWidget`
            (i.e. :guilabel:`IPython Console` tab),
            or if that kernel failed with an error, the error message widget.
        """
        return self._content_widget

    def get_focus_widget(self) -> QWidget:
        """
        Get the stack widget associated to the currently active shell widget.

        Used by
        :meth:`~spyder.api.widgets.main_widget.PluginMainWidget.change_visibility`
        when switching to the plugin.

        Returns
        -------
        QWidget
            The current widget in the stack, associated with the active
            :class:`~spyder.plugins.ipythonconsole.widgets.ShellWidget`
            (i.e. :guilabel:`IPython Console` tab), to give focus to.
        """
        return self._stack.currentWidget()

    # ---- SpyderWidgetMixin API
    # ------------------------------------------------------------------------
    def update_style(self) -> None:
        """
        Update the stylesheet and style of the stack widget.

        .. caution::

            If overriding this method, :class:`super` must be called
            for this widget to display correctly in the Spyder UI.

        Returns
        -------
        None
        """
        self._stack.setStyleSheet("QStackedWidget {padding: 0px; border: 0px}")

    # ---- Stack accesors
    # ------------------------------------------------------------------------
    def count(self) -> int:
        """
        Get the number of widgets in the stack.

        Returns
        -------
        int
            The number of widgets in the stack.
        """
        return self._stack.count()

    def get_widget_for_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> QWidget | None:
        """
        Retrieve the stacked widget corresponding to the given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to return the associated widget of.

        Returns
        -------
        QWidget | None
            The widget in the stack associated with ``shellwidget``,
            or ``None`` if not found.
        """
        shellwidget_id = id(shellwidget)
        if shellwidget_id in self._shellwidgets:
            return self._shellwidgets[shellwidget_id]
        return None

    # ---- Public API
    # ------------------------------------------------------------------------
    def add_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Create a new widget in the stack associated to a given shell widget.

        This method registers a new widget to display the content that
        is associated with the given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to associate the new widget to.

        Returns
        -------
        None
        """
        shellwidget_id = id(shellwidget)
        if shellwidget_id not in self._shellwidgets:
            widget = self.create_new_widget(shellwidget)
            self._stack.addWidget(widget)
            self._shellwidgets[shellwidget_id] = widget

            # Add all actions to new widget for shortcuts to work.
            for __, action in self.get_actions().items():
                if action:
                    widget_actions = widget.actions()
                    if action not in widget_actions:
                        widget.addAction(action)

            self.set_shellwidget(shellwidget)

    def remove_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Remove the stacked widget associated to a given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to remove the associated widget of.

        Returns
        -------
        None
        """
        shellwidget_id = id(shellwidget)
        if shellwidget_id in self._shellwidgets:
            widget = self._shellwidgets.pop(shellwidget_id)

            # If `widget` is an empty pane, we don't need to remove it from the
            # stack (because it's the one we need to show since the console is
            # showing an error) nor try to close it (because it makes no
            # sense).
            if not isinstance(widget, EmptyMessageWidget):
                self._stack.removeWidget(widget)
                self.close_widget(widget)

            self.update_actions()

    def set_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Set as active the stack widget associated with the given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget that corresponds to the stacked widget to set
            as currently active.

        Returns
        -------
        None
        """
        old_widget = self.current_widget()
        widget = self.get_widget_for_shellwidget(shellwidget)
        if widget is None:
            return

        self.set_content_widget(widget, add_to_stack=False)
        if (
            self.SHOW_MESSAGE_WHEN_EMPTY
            and self.get_conf(
                "show_message_when_panes_are_empty", section="main"
            )
            and widget.is_empty
        ):
            self.show_empty_message()
        else:
            self.show_content_widget()

        self.switch_widget(widget, old_widget)
        self.update_actions()

    def add_errored_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Add an error widget for a shell widget whose kernel failed to start.

        This is necessary to show a meaningful message when switching to
        consoles with dead kernels.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to associate with a new error widget.

        Returns
        -------
        None
        """
        shellwidget_id = id(shellwidget)

        # This can happen if the kernel started without issues but something is
        # printed to its stderr stream, which we display as an error in the
        # console. In that case, we need to remove the current widget
        # associated to shellwidget and replace it by an empty one.
        if shellwidget_id in self._shellwidgets:
            self._shellwidgets.pop(shellwidget_id)

        widget = _ErroredMessageWidget(self, shellwidget)
        widget.set_visibility(self.is_visible)
        if self.dockwidget is not None:
            self.dockwidget.visibilityChanged.connect(widget.set_visibility)

        self.set_content_widget(widget)
        self._shellwidgets[shellwidget_id] = widget
        self.set_shellwidget(shellwidget)

    def create_new_widget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> QWidget:
        """
        Create a new widget to communicate with the given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to create a new associated widget for.

        Returns
        -------
        QWidget
            The newly-created widget associated with ``shellwidget``.
        """
        raise NotImplementedError

    def close_widget(self, widget: QWidget) -> None:
        """
        Close the given stacked widget.

        Parameters
        ----------
        widget : QWidget
            The widget to close.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def switch_widget(self, widget: QWidget, old_widget: QWidget) -> None:
        """
        Switch the given stacked widget.

        Parameters
        ----------
        widget : QWidget
            The widget to switch to.
        old_widget : QWidget
            The previously-active widget.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def refresh(self) -> None:
        """
        Refresh the current stacked widget.

        Returns
        -------
        None
        """
        if self.count():
            widget = self.current_widget()
            widget.refresh()

    def is_current_widget_error_message(self) -> bool:
        """
        Check if the current widget is showing an error message.

        Returns
        -------
        bool
            ``True`` if the current widget is showing an error message,
            ``False`` if not.
        """
        return isinstance(self.current_widget(), _ErroredMessageWidget)

    def switch_empty_message(self, value: bool) -> None:
        """
        Switch between the empty message widget or the one with content.

        Parameters
        ----------
        value : bool
            If ``True``, switch to the empty widget; if ``False``, switch
            to the content widget.

        Returns
        -------
        None
        """
        if value:
            self.show_empty_message()
        else:
            self.show_content_widget()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""External terminal run executor configurations."""

# Standard library imports
import os.path as osp

# Third-party imports
from qtpy.compat import getopenfilename
from qtpy.QtCore import QSize
from qtpy.QtWidgets import (
    QCheckBox,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

# Local imports
from spyder.api.translations import _
from spyder.plugins.externalterminal.api import ExtTerminalShConfiguration
from spyder.plugins.run.api import (
    RunExecutorConfigurationGroup, Context, RunConfigurationMetadata,
    RunExecutorConfigurationGroupFactory)
from spyder.utils.icon_manager import ima
from spyder.utils.misc import getcwd_or_home
from spyder.utils.stylesheet import AppStyle


class ExternalTerminalPyConfiguration(RunExecutorConfigurationGroup):
    """External terminal Python run configuration options."""

    def __init__(self, parent, context: Context, input_extension: str,
                 input_metadata: RunConfigurationMetadata):
        super().__init__(parent, context, input_extension, input_metadata)

        self.dir = None

        # --- Interpreter ---
        interpreter_group = QGroupBox(_("Python interpreter"))
        interpreter_layout = QVBoxLayout(interpreter_group)

        # --- System terminal ---
        external_group = QWidget(self)
        external_layout = QGridLayout()
        external_group.setLayout(external_layout)

        self.interact_cb = QCheckBox(
            _("Interact with the interpreter after execution")
        )
        external_layout.addWidget(self.interact_cb, 1, 0, 1, -1)

        self.pclo_cb = QCheckBox(_("Interpreter options:"))
        external_layout.addWidget(self.pclo_cb, 3, 0)
        self.pclo_edit = QLineEdit(self)
        self.pclo_cb.toggled.connect(self.pclo_edit.setEnabled)
        self.pclo_edit.setEnabled(False)
        self.pclo_edit.setToolTip(
            _("<b>-u</b> is added to the other options you set here")
        )
        external_layout.addWidget(self.pclo_edit, 3, 1)

        interpreter_layout.addWidget(external_group)

        # --- General settings ----
        common_group = QGroupBox(_("Bash/Batch script settings"))
        common_layout = QGridLayout(common_group)

        self.clo_cb = QCheckBox(_("Command line options:"))
        common_layout.addWidget(self.clo_cb, 0, 0)
        self.clo_edit = QLineEdit(self)
        self.clo_edit.setMinimumWidth(300)
        self.clo_cb.toggled.connect(self.clo_edit.setEnabled)
        self.clo_edit.setEnabled(False)
        common_layout.addWidget(self.clo_edit, 0, 1)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(interpreter_group)
        layout.addWidget(common_group)
        layout.addStretch(100)

    @staticmethod
    def get_default_configuration() -> dict:
        return {
            'args_enabled': False,
            'args': '',
            'interact': True,
            'python_args_enabled': False,
            'python_args': '',
        }

    def set_configuration(self, config: dict):
        interact = config['interact']
        args_enabled = config['args_enabled']
        args = config['args']
        py_args_enabled = config['python_args_enabled']
        py_args = config['python_args']

        self.interact_cb.setChecked(interact)
        self.pclo_cb.setChecked(args_enabled)
        self.pclo_edit.setText(args)
        self.clo_cb.setChecked(py_args_enabled)
        self.clo_edit.setText(py_args)

    def get_configuration(self) -> dict:
        return {
            'args_enabled': self.pclo_cb.isChecked(),
            'args': self.pclo_edit.text(),
            'interact': self.interact_cb.isChecked(),
            'python_args_enabled': self.clo_cb.isChecked(),
            'python_args': self.clo_edit.text(),
        }


class GenericExternalTerminalShConfiguration(RunExecutorConfigurationGroup):
    """External terminal shell run configuration options."""

    def __init__(
        self,
        parent,
        context: Context, input_extension: str,
        input_metadata: RunConfigurationMetadata
    ):
        super().__init__(parent, context, input_extension, input_metadata)

        # --- Interpreter ---
        interpreter_group = QGroupBox(_("Interpreter"))
        interpreter_layout = QVBoxLayout(interpreter_group)

        interpreter_label = QLabel(_("Shell interpreter:"))
        self.interpreter_edit = QLineEdit(self)
        browse_btn = QPushButton(ima.icon('DirOpenIcon'), '', self)
        browse_btn.setToolTip(_("Select interpreter"))
        browse_btn.clicked.connect(self.select_interpreter)
        browse_btn.setIconSize(
            QSize(AppStyle.ConfigPageIconSize, AppStyle.ConfigPageIconSize)
        )

        shell_layout = QHBoxLayout()
        shell_layout.addWidget(interpreter_label)
        shell_layout.addWidget(self.interpreter_edit)
        shell_layout.addWidget(browse_btn)
        interpreter_layout.addLayout(shell_layout)

        self.interpreter_opts_cb = QCheckBox(_("Interpreter arguments:"))
        self.interpreter_opts_edit = QLineEdit(self)
        self.interpreter_opts_edit.setMinimumWidth(250)
        self.interpreter_opts_cb.toggled.connect(
            self.interpreter_opts_edit.setEnabled
        )
        self.interpreter_opts_edit.setEnabled(False)

        interpreter_opts_layout = QHBoxLayout()
        interpreter_opts_layout.addWidget(self.interpreter_opts_cb)
        interpreter_opts_layout.addWidget(self.interpreter_opts_edit)
        interpreter_layout.addLayout(interpreter_opts_layout)

        # --- Script ---
        script_group = QGroupBox(_('Script'))
        script_layout = QVBoxLayout(script_group)

        self.script_opts_cb = QCheckBox(_("Script arguments:"))
        self.script_opts_edit = QLineEdit(self)
        self.script_opts_cb.toggled.connect(
            self.script_opts_edit.setEnabled
        )
        self.script_opts_edit.setEnabled(False)

        script_args_layout = QHBoxLayout()
        script_args_layout.addWidget(self.script_opts_cb)
        script_args_layout.addWidget(self.script_opts_edit)
        script_layout.addLayout(script_args_layout)

        self.close_after_exec_cb = QCheckBox(
            _("Close terminal after execution")
        )
        script_layout.addWidget(self.close_after_exec_cb)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(interpreter_group)
        layout.addWidget(script_group)
        layout.addStretch(100)

    def select_interpreter(self):
        """Select an interpreter."""
        basedir = str(self.interpreter_edit.text())
        if not osp.isdir(basedir):
            basedir = getcwd_or_home()
        file, __ = getopenfilename(self, _("Select executable"), basedir)
        if file:
            self.interpreter_edit.setText(file)

    def set_configuration(self, config: ExtTerminalShConfiguration):
        interpreter = config['interpreter']
        interpreter_opts_enabled = config['interpreter_opts_enabled']
        interpreter_opts = config['interpreter_opts']
        script_opts_enabled = config['script_opts_enabled']
        script_opts = config['script_opts']
        close_after_exec = config['close_after_exec']

        self.interpreter_edit.setText(interpreter)
        self.interpreter_opts_cb.setChecked(interpreter_opts_enabled)
        self.interpreter_opts_edit.setText(interpreter_opts)
        self.script_opts_cb.setChecked(script_opts_enabled)
        self.script_opts_edit.setText(script_opts)
        self.close_after_exec_cb.setChecked(close_after_exec)

    def get_configuration(self) -> ExtTerminalShConfiguration:
        return {
            'interpreter': self.interpreter_edit.text(),
            'interpreter_opts_enabled': self.interpreter_opts_cb.isChecked(),
            'interpreter_opts': self.interpreter_opts_edit.text(),
            'script_opts_enabled': self.script_opts_cb.isChecked(),
            'script_opts': self.script_opts_edit.text(),
            'close_after_exec': self.close_after_exec_cb.isChecked()
        }


class MetaShConfiguration(type(GenericExternalTerminalShConfiguration)):
    def __new__(cls, clsname, bases, attrs):
        interp = attrs.pop('default_shell_meta')
        interp_opts = attrs.pop('shell_args_meta')
        interp_opts_enabled = interp_opts != ''

        def get_default_conf() -> RunExecutorConfigurationGroupFactory:
            return {
                'interpreter': interp,
                'interpreter_opts_enabled': interp_opts_enabled,
                'interpreter_opts': interp_opts,
                'script_opts_enabled': False,
                'script_opts': '',
                'close_after_exec': False,
            }

        return super(MetaShConfiguration, cls).__new__(cls, clsname, bases, {
            **attrs,
            'get_default_configuration': staticmethod(get_default_conf)
        })


def ExternalTerminalShConfiguration(
    default_shell: str,
    shell_args: str = ''
) -> RunExecutorConfigurationGroup:

    class WrappedExternalTerminalShConfiguration(
        GenericExternalTerminalShConfiguration,
        metaclass=MetaShConfiguration
    ):
        default_shell_meta = default_shell
        shell_args_meta = shell_args

    return WrappedExternalTerminalShConfiguration



# -----------------------------------------------------------------------------
# Copyright (c) 2020- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
The API-related exception classes emitted by Spyder and its plugins.
"""


class SpyderAPIError(Exception):
    """An error due to incorrect usage by the consumer of the Spyder API."""

    pass



# -----------------------------------------------------------------------------
# Copyright (c) 2020- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
The gettext-based translation and internationalization (i18n) API for Spyder.
"""

from spyder.config.base import get_translation, _


__all__ = ["get_translation", "_"]
